"""Google Colab utilities."""
import os

from hyfi.utils.env import set_osenv
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def mount_google_drive(
    project_root: str = "",
    project_name: str = "",
    mountpoint: str = "/content/drive",
    force_remount: bool = False,
    timeout_ms: int = 120000,
) -> None:
    """Mount Google Drive to Colab."""
    try:
        from google.colab import drive  # type: ignore

        drive.mount(mountpoint, force_remount=force_remount, timeout_ms=timeout_ms)

        if project_root:
            if not project_root.startswith(os.path.sep) and not project_root.startswith(
                ".."
            ):
                project_root = os.path.join(mountpoint, project_root)
            set_osenv("HYFI_PROJECT_ROOT", project_root)
            logger.info(f"Setting HYFI_PROJECT_ROOT to {project_root}")
        if project_name:
            set_osenv("HYFI_PROJECT_NAME", project_name)
            logger.info(f"Setting HYFI_PROJECT_NAME to {project_name}")
    except ImportError:
        logger.warning("Google Colab not detected.")



"""Some utility functions."""

import errno
import os
import shutil
import stat
import sys
import tempfile
import warnings
from contextlib import suppress
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, Optional, TextIO, Tuple, Union

import colorama
from pydantic import StrictBool

from hyfi.utils.types import IntSeq

colorama.init()


class Style:
    """Common color styles."""

    OK: IntSeq = [colorama.Fore.GREEN, colorama.Style.BRIGHT]
    WARNING: IntSeq = [colorama.Fore.YELLOW, colorama.Style.BRIGHT]
    IGNORE: IntSeq = [colorama.Fore.CYAN]
    DANGER: IntSeq = [colorama.Fore.RED, colorama.Style.BRIGHT]
    RESET: IntSeq = [colorama.Fore.RESET, colorama.Style.RESET_ALL]


INDENT = " " * 2
HLINE = "-" * 42

NO_VALUE: object = object()


def printf(
    action: str,
    msg: Any = "",
    style: Optional[IntSeq] = None,
    indent: int = 10,
    verbose: Union[bool, StrictBool] = True,
    file_: TextIO = sys.stdout,
) -> Optional[str]:
    """Print string with common format."""
    if not verbose:
        return None  # HACK: Satisfy MyPy
    _msg = str(msg)
    action = action.rjust(indent, " ")
    if not style:
        return action + _msg

    out = style + [action] + Style.RESET + [INDENT, _msg]
    print(*out, sep="", file=file_)
    return None  # HACK: Satisfy MyPy


def printf_exception(
    e: Exception, action: str, msg: str = "", indent: int = 0, quiet: bool = False
) -> None:
    """Print exception with common format."""
    if not quiet:
        print("", file=sys.stderr)
        printf(action, msg=msg, style=Style.DANGER, indent=indent, file_=sys.stderr)
        print(HLINE, file=sys.stderr)
        print(e, file=sys.stderr)
        print(HLINE, file=sys.stderr)


def cast_str_to_bool(value: Any) -> bool:
    """Parse anything to bool.

    Params:
        value:
            Anything to be casted to a bool. Tries to be as smart as possible.

            1.  Cast to number. Then: 0 = False; anything else = True.
            1.  Find [YAML booleans](https://yaml.org/type/bool.html),
                [YAML nulls](https://yaml.org/type/null.html) or `none` in it
                and use it appropriately.
            1.  Cast to boolean using standard python `bool(value)`.
    """
    # Assume it's a number
    with suppress(TypeError, ValueError):
        return bool(float(value))
    # Assume it's a string
    with suppress(AttributeError):
        lower = value.lower()
        if lower in {"y", "yes", "t", "true", "on"}:
            return True
        elif lower in {"n", "no", "f", "false", "off", "~", "null", "none"}:
            return False
    # Assume nothing
    return bool(value)


def copy_file(src_path: Path, dst_path: Path, follow_symlinks: bool = True) -> None:
    """Copy one file to another place."""
    shutil.copy2(src_path, dst_path, follow_symlinks=follow_symlinks)


def force_str_end(original_str: str, end: str = "\n") -> str:
    """Make sure a `original_str` ends with `end`.

    Params:
        original_str: String that you want to ensure ending.
        end: String that must exist at the end of `original_str`
    """
    return original_str + end if not original_str.endswith(end) else original_str


def handle_remove_readonly(
    func: Callable, path: str, exc: Tuple[BaseException, OSError, TracebackType]
) -> None:
    """Handle errors when trying to remove read-only files through `shutil.rmtree`.

    This handler makes sure the given file is writable, then re-execute the given removal function.

    Arguments:
        func: An OS-dependant function used to remove a file.
        path: The path to the file to remove.
        exc: A `sys.exc_info()` object.
    """
    excvalue = exc[1]
    if func in (os.rmdir, os.remove, os.unlink) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


# See https://github.com/copier-org/copier/issues/345
class TemporaryDirectory(tempfile.TemporaryDirectory):
    """A custom version of `tempfile.TemporaryDirectory` that handles read-only files better.

    On Windows, before Python 3.8, `shutil.rmtree` does not handle read-only files very well.
    This custom class makes use of a [special error handler][copier.tools.handle_remove_readonly]
    to make sure that a temporary directory containing read-only files (typically created
    when git-cloning a repository) is properly cleaned-up (i.e. removed) after using it
    in a context manager.
    """

    @classmethod
    def _cleanup(cls, name, warn_message):
        cls._robust_cleanup(name)
        warnings.warn(warn_message, ResourceWarning)

    def cleanup(self):
        """Remove directory safely."""
        if self._finalizer.detach():
            self._robust_cleanup(self.name)

    @staticmethod
    def _robust_cleanup(name):
        shutil.rmtree(name, ignore_errors=False, onerror=handle_remove_readonly)


def readlink(link: Path) -> Path:
    """A custom version of os.readlink/pathlib.Path.readlink.

    pathlib.Path.readlink is what we ideally would want to use, but it is only available on python>=3.9.
    os.readlink doesn't support Path and bytes on Windows for python<3.8
    """
    if sys.version_info >= (3, 9):
        return link.readlink()
    elif sys.version_info >= (3, 8) or os.name != "nt":
        return Path(os.readlink(link))
    else:
        return Path(os.readlink(str(link)))



from typing import Any

from pydantic import BaseModel

from hyfi.__global__ import __about__
from hyfi.hydra import _compose
from hyfi.joblib.batch import batcher
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class DistFramwork(BaseModel):
    """Distributed Framework Configuration"""

    backend: str = "joblib"
    initialize: bool = False
    num_workers: int = 1


class BatcherConfig(BaseModel):
    """Batcher Configuration"""

    procs: int = 1
    minibatch_size: int = 1_000
    backend: str = "joblib"
    task_num_cpus: int = 1
    task_num_gpus: int = 0
    verbose: int = 10


class JobLibConfig(BaseModel):
    """JobLib Configuration"""

    config_name: str = "__init__"
    num_workers: int = 1
    distributed_framework: DistFramwork = DistFramwork()
    batcher: BatcherConfig = BatcherConfig()
    __initilized__: bool = False

    class Config:
        extra = "allow"
        underscore_attrs_are_private = True

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        data = _compose(
            f"joblib={config_name}",
            config_data=data,
            config_module=__about__.config_module,
        )  # type: ignore
        super().__init__(config_name=config_name, **data)

    def init_backend(
        self,
    ):
        """Initialize the backend for joblib"""
        if self.distributed_framework.initialize:
            backend_handle = None
            backend = self.distributed_framework.backend

            if backend == "dask":
                from dask.distributed import Client  # type: ignore

                dask_cfg = {"n_workers": self.distributed_framework.num_workers}
                logger.info(f"initializing dask client with {dask_cfg}")
                client = Client(**dask_cfg)
                logger.debug(client)

            elif backend == "ray":
                import ray  # type: ignore

                ray_cfg = {"num_cpus": self.distributed_framework.num_workers}
                logger.info(f"initializing ray with {ray_cfg}")
                ray.init(**ray_cfg)
                backend_handle = ray

            batcher.batcher_instance = batcher.Batcher(
                backend_handle=backend_handle, **self.batcher.dict()
            )
            logger.info(f"initialized batcher with {batcher.batcher_instance}")
        self.__initilized__ = True

    def stop_backend(self):
        """Stop the backend for joblib"""
        backend = self.distributed_framework.backend
        if batcher.batcher_instance:
            logger.info("stopping batcher")
            del batcher.batcher_instance

        logger.info("stopping distributed framework")
        if self.distributed_framework.initialize:
            if backend == "ray":
                try:
                    import ray  # type: ignore

                    if ray.is_initialized():
                        ray.shutdown()
                        logger.info("shutting down ray")
                except ImportError:
                    logger.warning("ray is not installed")

            elif backend == "dask":
                try:
                    from dask.distributed import Client  # type: ignore

                    if Client.initialized():
                        Client.close()
                        logger.info("shutting down dask client")
                except ImportError:
                    logger.warning("dask is not installed")



"""Image utils."""
import io

import numpy as np
from PIL import Image, ImageFont

from hyfi.graphic.plot import get_plot_font
from hyfi.utils.file import read


def scale_image(
    image: Image.Image,
    max_width: int = None,
    max_height: int = None,
    max_pixels: int = None,
    scale: float = 1.0,
    resize_to_multiple_of: int = 8,
    resample: int = Image.LANCZOS,
) -> Image.Image:
    """Scale image to have at most `max_pixels` pixels."""
    w, h = image.size

    if max_width is None and max_height is not None:
        max_width = int(w * max_height / h)
    elif max_height is None and max_width is not None:
        max_height = int(h * max_width / w)

    if max_width is not None and max_height is not None:
        max_pixels = max_width * max_height
    if max_pixels is not None:
        scale = np.sqrt(max_pixels / (w * h))

    max_width = int(w * scale)
    max_height = int(h * scale)
    if resize_to_multiple_of is not None:
        max_width = (max_width // resize_to_multiple_of) * resize_to_multiple_of
        max_height = (max_height // resize_to_multiple_of) * resize_to_multiple_of

    if scale < 1.0 or w > max_width or h > max_height:
        image = image.resize((max_width, max_height), resample=resample)
    return image


def load_image(
    image_or_uri,
    max_width: int = None,
    max_height: int = None,
    max_pixels: int = None,
    scale: float = 1.0,
    resize_to_multiple_of: int = None,
    crop_box=None,
    mode="RGB",
    **kwargs
) -> Image.Image:
    """Load image from file or URI."""
    from PIL import Image

    if isinstance(image_or_uri, Image.Image):
        img = image_or_uri.convert(mode)
    else:
        img = Image.open(io.BytesIO(read(image_or_uri, **kwargs))).convert(mode)
    img = scale_image(
        img,
        max_width=max_width,
        max_height=max_height,
        max_pixels=max_pixels,
        scale=scale,
        resize_to_multiple_of=resize_to_multiple_of,
    )
    if crop_box is not None:
        img = img.crop(crop_box)
    return img


def load_images(
    images_or_uris,
    max_width=None,
    max_height=None,
    max_pixels=None,
    scale=1.0,
    resize_to_multiple_of: int = None,
    crop_to_min_size=False,
    mode="RGB",
    **kwargs
):
    """Load images from files or URIs."""
    imgs = [
        load_image(
            image_or_uri,
            max_width=max_width,
            max_height=max_height,
            max_pixels=max_pixels,
            scale=scale,
            resize_to_multiple_of=resize_to_multiple_of,
            mode=mode,
            **kwargs
        )
        for image_or_uri in images_or_uris
    ]
    if crop_to_min_size:
        min_width = min(img.width for img in imgs)
        min_height = min(img.height for img in imgs)
        if resize_to_multiple_of is not None:
            min_width = (min_width // resize_to_multiple_of) * resize_to_multiple_of
            min_height = (min_height // resize_to_multiple_of) * resize_to_multiple_of
        imgs = [img.crop((0, 0, min_width, min_height)) for img in imgs]

    return imgs


def get_image_font(fontname=None, fontsize=12):
    """Get font for PIL image."""
    fontname, fontpath = get_plot_font(set_font_for_matplot=False, fontname=fontname)
    return ImageFont.truetype(fontpath, fontsize) if fontpath else None



"""GPU utilities"""

import contextlib
import gc
import os
import time
from threading import Thread

from hyfi.utils.logging import getLogger
from hyfi.utils.notebook import clear_output

logger = getLogger(__name__)

try:
    import GPUtil  # type: ignore
except ImportError:
    logger.info("GPUtil not found. Please install it to use GPU utilities.")


class GPUMon(Thread):
    """Monitor GPU usage in a separate thread"""

    def __init__(self, delay=10, show_current_time=True, clear_output=True):
        super(GPUMon, self).__init__()
        self.stopped = False
        self.delay = delay  # Time between calls to GPUtil
        self.show_current_time = show_current_time
        self.clear_output = clear_output
        self.start()

    def run(self):
        while not self.stopped:
            if self.clear_output:
                clear_output()
            if self.show_current_time:
                print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
            GPUtil.showUtilization()
            time.sleep(self.delay)

    def stop(self):
        self.stopped = True

    @staticmethod
    def gpu_usage(all=False, attrList=None, useOldCode=False):
        return GPUtil.showUtilization(all=all, attrList=attrList, useOldCode=useOldCode)

    @staticmethod
    def get_gpus():
        return GPUtil.getGPUs()

    @staticmethod
    def get_gpu_info():
        gpus = GPUtil.getGPUs()
        return [
            {
                "id": gpu.id,
                "name": gpu.name,
                "load": gpu.load,
                "memory_used": gpu.memoryUsed,
                "memory_total": gpu.memoryTotal,
                "temperature": gpu.temperature,
            }
            for gpu in gpus
        ]

    @staticmethod
    def get_first_available(
        order="first",
        maxLoad=0.5,
        maxMemory=0.5,
        attempts=1,
        interval=900,
        verbose=False,
        includeNan=False,
        excludeID=[],
        excludeUUID=[],
    ):
        return GPUtil.getFirstAvailable(
            order=order,
            maxLoad=maxLoad,
            maxMemory=maxMemory,
            attempts=attempts,
            interval=interval,
            verbose=verbose,
            includeNan=includeNan,
            excludeID=excludeID,
            excludeUUID=excludeUUID,
        )

    @staticmethod
    def release_gpu_memory():
        gc.collect()
        with contextlib.suppress(ImportError):
            import torch  # type: ignore

            torch.cuda.empty_cache()


def nvidia_smi():
    """Run nvidia-smi and return the output as a string"""
    import subprocess

    return subprocess.run(["nvidia-smi", "-L"], stdout=subprocess.PIPE).stdout.decode(
        "utf-8"
    )


def is_cuda_available():
    """Check if cuda is available"""
    try:
        import torch  # type: ignore

        return torch.cuda.is_available()
    except ImportError:
        return False


def set_cuda(device=0):
    """Set cuda device to use"""
    try:
        import torch  # type: ignore

        _names = []
        if isinstance(device, str):
            device = device.replace("cuda:", "")
            ids = device.split(",")
        else:
            ids = [str(device)]
        for id in ids:
            _device_name = torch.cuda.get_device_name(int(id))
            _names.append(f"{_device_name} (id:{id})")
        logger.info(f"Setting cuda device to {_names}")
        device = ", ".join(ids)
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = device
    except ImportError as e:
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
        raise Exception("Cuda device not found") from e



from pydantic import BaseModel

from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class ModuleConfig(BaseModel):
    """Module Configuration"""

    library_dir: str = ""
    modules: list = []



"""Logging utilities"""
import logging
import os
import sys
import warnings


def setLogger(level=None, force=True, filterwarnings_action="ignore", **kwargs) -> None:
    """
    Set the logging level and format. This is a wrapper around logging. basicConfig to allow the user to specify a different logging level for each test and to filter warnings that are not caught by the filterwarnings_action

    Args:
        level: The logging level to use
        force: If True ( default ) all warnings will be logged even if there are no warnings
        filterwarnings_action: The action to call when a warning is logged

    Returns:
        The logger that was
    """
    level = level or os.environ.get("HYFI_LOG_LEVEL") or "INFO"
    level = level.upper()
    os.environ["HYFI_LOG_LEVEL"] = level
    # Filter warnings by applying filterwarnings_action to the warnings.
    if filterwarnings_action is not None:
        warnings.filterwarnings(filterwarnings_action)
    # Return the logging level.
    if isinstance(level, str):
        level = getattr(logging, level.upper(), logging.INFO)
    # Configure logging level level and force logging.
    if sys.version_info >= (3, 8):
        logging.basicConfig(level=level, force=force, **kwargs)
    else:
        logging.basicConfig(level=level, **kwargs)


def getLogger(
    _name=None,
    _log_level=None,
    **kwargs,
) -> logging.Logger:
    """
    Get a logger with a given name and log level. It is possible to pass a logger name and log level to this function.

    Args:
        _name: The name of the logger to get. If not specified the name of the module is used.
        _log_level: The log level to set.

    Returns:
        The logger with the given name and log level set to the value specified in HYFI_LOG_LEVEL
    """
    _name = _name or __name__
    logger = logging.getLogger(_name)
    _log_level = _log_level or os.environ.get("HYFI_LOG_LEVEL") or "INFO"
    logger.setLevel(_log_level)
    return logger



"""Batcher class for handling parallel jobs on minibatches"""
import contextlib
import multiprocessing
import random
from contextlib import closing
from math import ceil

import pandas as pd
import scipy.sparse as ssp
from tqdm.auto import tqdm

from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class Batcher(object):
    """Scheduler to handle parallel jobs on minibatches

    Parameters
    ----------
    procs: int
            Number of process(es)/thread(s) for executing task in parallel. Used for multiprocessing, threading and Loky

    minibatch_size: int
            Expected size of each minibatch

    backend: {'serial', 'multiprocessing', 'threading', 'loky', 'spark', 'dask', 'ray'}
            Backend for computing the tasks

                    - 'serial' sequential execution without a backend scheduler

                    - 'multiprocessing' Python standard multiprocessing library

                    - 'threading' Python standard threading library

                    - 'loky' Loky fork of multiprocessing library

                    - 'joblib' Joblib fork of multiprocessing library

                    - 'spark' PySpark local or distributed execution

                    - 'dask' Dask Distributed local or distributed execution

                    - 'ray' Ray local or distributed execution

    task_num_cpus: int
            Number of CPUs to reserve per minibatch task for Ray

    task_num_gpus: int
            Number of GPUs to reserve per minibatch task for Ray

    backend_handle: object
            Backend handle for sending tasks

    verbose: int
            Verbosity level.
            Setting verbose > 0 will display additional information depending on the specific level set.
    """

    def __init__(
        self,
        procs=0,
        minibatch_size=20000,
        backend_handle=None,
        backend="multiprocessing",
        task_num_cpus=1,
        task_num_gpus=0,
        verbose=0,
    ):
        if isinstance(procs, str):
            procs = int(procs)
        if procs == 0 or procs is None:
            procs = multiprocessing.cpu_count()
        self.procs = procs
        self.verbose = verbose
        self.minibatch_size = minibatch_size
        self.backend_handle = backend_handle
        self.backend = backend
        self.task_num_cpus = task_num_cpus
        self.task_num_gpus = task_num_gpus

    def list2indexedrdd(self, lst, minibatch_size=0):
        if minibatch_size == 0:
            minibatch_size = self.minibatch_size
        start = 0
        len_data = len(lst)
        batch_count = 0
        batches = []
        while start < len_data:
            batches.append([batch_count] + [lst[start : start + minibatch_size]])
            start += minibatch_size
            batch_count += 1
        return self.backend_handle.parallelize(batches)

    def indexedrdd2list(self, indexedrdd, sort=True):
        batches = indexedrdd.collect()
        if sort:
            batches = sorted(batches)
        return [batch[1] for batch in batches]

    def split_batches(self, data, minibatch_size=None, backend=None):
        """Split data into minibatches with a specified size

        Parameters
        ----------
        data: iterable and indexable
                List-like data to be split into batches. Includes backend_handleipy matrices and Pandas DataFrames.

        minibatch_size: int
                Expected sizes of minibatches split from the data.

        backend: object
                Backend to use, instead of the Batcher backend attribute

        Returns
        -------
        data_split: list
                List of minibatches, each entry is a list-like object representing the data subset in a batch.
        """
        if minibatch_size is None:
            minibatch_size = self.minibatch_size
        if backend is None:
            backend = self.backend
        if isinstance(data, (list, tuple, dict)):
            len_data = len(data)
        else:
            len_data = data.shape[0]
        if backend == "spark":
            return self.list2indexedrdd(data, minibatch_size)
        if isinstance(data, pd.DataFrame):
            data = [
                data.iloc[x * minibatch_size : (x + 1) * minibatch_size]
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        elif isinstance(data, dict):
            data = [
                dict(
                    list(data.items())[
                        x * minibatch_size : min(len_data, (x + 1) * minibatch_size)
                    ]
                )
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        else:
            data = [
                data[x * minibatch_size : min(len_data, (x + 1) * minibatch_size)]
                for x in range(int(ceil(len_data / minibatch_size)))
            ]
        # if backend=="dask":  return self.backend_handle.scatter(data)
        return data

    def collect_batches(self, data, backend=None, sort=True):
        if backend is None:
            backend = self.backend
        if backend == "spark":
            data = self.indexedrdd2list(data, sort)
        if backend == "dask":
            data = self.backend_handle.gather(data)
        return data

    def merge_batches(self, data):
        """Merge a list of data minibatches into one single instance representing the data

        Parameters
        ----------
        data: list
                List of minibatches to merge

        Returns
        -------
        (anonymous): sparse matrix | pd.DataFrame | list
                Single complete list-like data merged from given batches
        """
        if isinstance(data[0], ssp.csr_matrix):
            return ssp.vstack(data)
        if isinstance(data[0], (pd.DataFrame, pd.Series)):
            return pd.concat(data)
        return [item for sublist in data for item in sublist]

    def process_batches(
        self,
        task,
        data,
        args,
        backend=None,
        backend_handle=None,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        procs=None,
        task_num_cpus=None,
        task_num_gpus=None,
        verbose=None,
        description="batch_apply",
    ):
        """

        Parameters
        ----------
        task: function
                Function to apply on each minibatch with other specified arguments

        data: list-like
                Samples to split into minibatches and apply the specified function on

        args: list
                Arguments to pass to the specified function following the mini-batch

        input_split: boolean, default False
                If True, input data is already mapped into minibatches, otherwise data will be split on call.

        merge_output: boolean, default True
                If True, results from minibatches will be reduced into one single instance before return.

        procs: int
                Number of process(es)/thread(s) for executing task in parallel. Used for multiprocessing, threading,
                Loky and Ray

        minibatch_size: int
                Expected size of each minibatch

        backend: {'serial', 'multiprocessing', 'threading', 'loky', 'spark', 'dask', 'ray'}
                Backend for computing the tasks

                        - 'serial' sequential execution without a backend scheduler

                        - 'multiprocessing' Python standard multiprocessing library

                        - 'threading' Python standard threading library

                        - 'loky' Loky fork of multiprocessing library

                        - 'joblib' Joblib fork of multiprocessing library

                        - 'spark' PySpark local or distributed execution

                        - 'dask' Dask Distributed local or distributed execution

                        - 'ray' Ray local or distributed execution

        backend_handle: object
                Backend handle for sending tasks

        task_num_cpus: int
                Number of CPUs to reserve per minibatch task for Ray

        task_num_gpus: int
                Number of GPUs to reserve per minibatch task for Ray

        verbose: int
                Verbosity level.
                Setting verbose > 0 will display additional information depending on the specific level set.

        Returns
        -------
        results: list-like | list of list-like
                If merge_output is specified as True, this will be a list-like object representing
                the dataset, with each entry as a sample. Otherwise this will be a list of list-like
                objects, with each entry representing the results from a minibatch.
        """
        if procs is None:
            procs = self.procs
        if backend is None:
            backend = self.backend
        if backend_handle is None:
            backend_handle = self.backend_handle
        if task_num_cpus is None:
            task_num_cpus = self.task_num_cpus
        if task_num_gpus is None:
            task_num_gpus = self.task_num_gpus
        if verbose is None:
            verbose = self.verbose
        if verbose > 1:
            logger.info(
                f" backend: {backend}  minibatch_size: {self.minibatch_size}  procs: {procs}  input_split: {input_split}  merge_output: {merge_output}  len(data): {len(data)} len(args): {len(args)}"
            )

        if verbose > 10:
            logger.info(
                f" len(data): {len(data)} len(args): {len(args)} [type(x) for x in data]: {[type(x) for x in data]} [type(x) for x in args]: {[type(x) for x in args]}"
            )

        if not (input_split):
            if backend == "spark":
                paral_params = self.split_batches(data, minibatch_size, backend="spark")
            else:
                paral_params = [
                    [data_batch] + args
                    for data_batch in self.split_batches(data, minibatch_size)
                ]
        else:
            if backend != "spark":
                paral_params = [[data_batch] + args for data_batch in data]
            else:
                paral_params = data
        if verbose > 10:
            print(" Start task, len(paral_params)", len(paral_params))
        if backend == "serial":
            results = [
                task(minibatch) for minibatch in tqdm(paral_params, desc=description)
            ]
        else:
            if backend == "multiprocessing":
                with closing(
                    multiprocessing.Pool(max(1, procs), maxtasksperchild=2)
                ) as pool:
                    results = pool.map_async(task, paral_params)
                    pool.close()
                    pool.join()
                    results = results.get()
            elif backend == "threading":
                with closing(multiprocessing.dummy.Pool(max(1, procs))) as pool:
                    results = pool.map(task, paral_params)
                    pool.close()
                    pool.join()
            elif backend == "loky":
                from loky import get_reusable_executor

                pool = get_reusable_executor(max_workers=max(1, procs))
                results = list(pool.map(task, tqdm(paral_params, desc=description)))
            elif backend == "joblib":
                from joblib import Parallel, delayed

                with tqdm_joblib(
                    tqdm(desc=description, total=len(paral_params))
                ) as pbar:
                    results = Parallel(n_jobs=procs)(
                        delayed(task)(params) for params in paral_params
                    )
            elif backend == "p_tqdm":
                from p_tqdm import p_map

                results = p_map(task, paral_params, num_cpus=procs)
            elif backend == "dask":
                # if not (input_split):  data= self.scatter(data)
                results = [
                    self.backend_handle.submit(task, params)
                    for params in tqdm(paral_params, desc=description)
                ]
            elif backend == "spark":

                def apply_func_to_indexedrdd(batch):
                    return [batch[0]] + [task([batch[1]] + args)]

                results = paral_params.map(apply_func_to_indexedrdd)
            elif backend == "ray":

                @self.backend_handle.remote(
                    num_cpus=task_num_cpus, num_gpus=task_num_gpus
                )
                def f_ray(f, data):
                    return f(data)

                pbar = tqdm(desc=description, total=len(paral_params) + 1)
                results = [
                    f_ray.remote(task, paral_params.pop(0))
                    for _ in range(min(len(paral_params), self.procs))
                ]
                uncompleted = results
                pbar.update(len(results))
                while len(paral_params) > 0:
                    # More tasks than available processors. Queue the task calls
                    done, remaining = self.backend_handle.wait(
                        uncompleted, timeout=60, fetch_local=False
                    )
                    # if verbose > 5:  print("Done, len(done), len(remaining)", len(done), len(remaining))
                    if len(done) == 0:
                        continue
                    done = done[0]
                    uncompleted = [x for x in uncompleted if x != done]
                    if len(remaining) > 0:
                        new = f_ray.remote(task, paral_params.pop(0))
                        pbar.update(1)
                        uncompleted.append(new)
                        results.append(new)
                results = [self.backend_handle.get(x) for x in results]
                pbar.update(1)
                pbar.close()
            # ppft currently not supported. Supporting arbitrary tasks requires modifications to passed arguments
            # elif backend == "ppft":
            #   jobs = [self.backend_handle.submit(task, (x,), (), ()) for x in paral_params]
            # 	results = [x() for x in jobs]

        if merge_output:
            return self.merge_batches(self.collect_batches(results, backend=backend))
        if verbose > 2:
            logger.info(
                f" Task: {task}  backend: {backend}  backend_handle: {backend_handle}  completed"
            )
        return results

    def shuffle_batch(self, texts, labels=None, seed=None):
        """Shuffle a list of samples, as well as the labels if specified

        Parameters
        ----------
        texts: list-like
                List of samples to shuffle

        labels: list-like (optional)
                List of labels to shuffle, should be correspondent to the samples given

        seed: int
                The seed of the pseudo random number generator to use for shuffling

        Returns
        -------
        texts: list
                List of shuffled samples (texts parameters)

        labels: list (optional)
                List of shuffled labels. This will only be returned when non-None
                labels is passed
        """
        if seed is not None:
            random.seed(seed)
        index_shuf = list(range(len(texts)))
        random.shuffle(index_shuf)
        texts = [texts[x] for x in index_shuf]
        if labels is None:
            return texts
        labels = [labels[x] for x in index_shuf]
        return texts, labels

    def __getstate__(self):
        return dict(self.__dict__.items())

    def __setstate__(self, params):
        for key in params:
            setattr(self, key, params[key])


batcher_instance: Batcher


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    import joblib

    class TqdmBatchCompletionCallBack(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallBack
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


# from joblib import Parallel, delayed

# with tqdm_joblib(tqdm(desc="My calculation", total=10)) as progress_bar:
#     Parallel(n_jobs=16)(delayed(sqrt)(i**2) for i in range(10))



import functools
import importlib
import inspect
import json
import os
import random
from enum import Enum
from pathlib import Path
from typing import IO, Any, Dict, List, Tuple, Union

import hydra
from omegaconf import DictConfig, ListConfig, OmegaConf, SCMode

from hyfi.__global__ import __home_path__, __hyfi_path__
from hyfi.__global__.config import __global_config__
from hyfi.cached_path import cached_path
from hyfi.hydra import _compose, _select, _to_dict
from hyfi.utils.env import dotenv_values, getcwd
from hyfi.utils.file import check_path, exists, join_path, mkdir
from hyfi.utils.func import lower_case_with_underscores, strptime, today
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


DictKeyType = Union[str, int, Enum, float, bool]


class SpecialKeys(str, Enum):
    """Special keys in configs used by hyfi."""

    CALL = "_call_"
    CONFIG = "_config_"
    CONFIG_GROUP = "_config_group_"
    EXEC = "_exec_"
    FUNC = "_func_"
    METHOD = "_method_"
    METHOD_NAME = "_name_"
    NAME = "name"
    PARTIAL = "_partial_"
    rcPARAMS = "rcParams"
    RECURSIVE = "_recursive_"
    SUFFIX = "suffix"
    TARGET = "_target_"
    VERBOSE = "verbose"


_config_ = _compose().copy()


def _print(cfg: Any, resolve: bool = True, **kwargs):
    import pprint

    if _is_config(cfg):
        if resolve:
            pprint.pprint(_to_dict(cfg), **kwargs)
        else:
            pprint.pprint(cfg, **kwargs)
    else:
        print(cfg)


def _is_config(
    cfg: Any,
):
    return isinstance(cfg, (DictConfig, dict))


def _is_list(
    cfg: Any,
):
    return isinstance(cfg, (ListConfig, list))


def _is_instantiatable(cfg: Any):
    return _is_config(cfg) and SpecialKeys.TARGET in cfg


def _load(file_: Union[str, Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
    return OmegaConf.load(file_)


def _save(config: Any, f: Union[str, Path, IO[Any]], resolve: bool = False) -> None:
    os.makedirs(os.path.dirname(str(f)), exist_ok=True)
    OmegaConf.save(config, f, resolve=resolve)


def _save_json(
    json_dict: dict,
    f: Union[str, Path, IO[Any]],
    indent=4,
    ensure_ascii=False,
    default=None,
    **kwargs,
):
    f = str(f)
    os.makedirs(os.path.dirname(f), exist_ok=True)
    with open(f, "w") as f:
        json.dump(
            json_dict,
            f,
            indent=indent,
            ensure_ascii=ensure_ascii,
            default=default,
            **kwargs,
        )


def _load_json(f: Union[str, Path, IO[Any]], **kwargs) -> dict:
    f = str(f)
    with open(f, "r") as f:
        return json.load(f, **kwargs)


def _update(_dict, _overrides):
    import collections.abc

    for k, v in _overrides.items():
        if isinstance(v, collections.abc.Mapping):
            _dict[k] = _update((_dict.get(k) or {}), v)
        else:
            _dict[k] = v
    return _dict


def _merge(
    *configs: Union[
        DictConfig,
        ListConfig,
        Dict[DictKeyType, Any],
        List[Any],
        Tuple[Any, ...],
        Any,
    ],
) -> Union[ListConfig, DictConfig]:
    """
    Merge a list of previously created configs into a single one
    :param configs: Input configs
    :return: the merged config object.
    """
    return OmegaConf.merge(*configs)


def _to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
    return OmegaConf.to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)


def _to_container(
    cfg: Any,
    *,
    resolve: bool = False,
    throw_on_missing: bool = False,
    enum_to_str: bool = False,
    structured_config_mode: SCMode = SCMode.DICT,
):
    return OmegaConf.to_container(
        cfg,
        resolve=resolve,
        throw_on_missing=throw_on_missing,
        enum_to_str=enum_to_str,
        structured_config_mode=structured_config_mode,
    )


def _run(config: Any, **kwargs: Any) -> Any:
    config = _merge(config, kwargs)
    _config_ = config.get(SpecialKeys.CONFIG)
    if _config_ is None:
        logger.warning("No _config_ specified in config")
        return None
    if isinstance(_config_, str):
        _config_ = [_config_]
    for _cfg_ in _config_:
        cfg = _select(config, _cfg_)
        _instantiate(cfg)


def _partial(
    config: Any = None, config_group: Union[str, None] = None, *args: Any, **kwargs: Any
) -> Any:
    if config is None and config_group is None:
        logger.warning("No config specified")
        return None
    elif config_group is not None:
        config = _compose(config_group=config_group)
    kwargs[SpecialKeys.PARTIAL] = True
    return _instantiate(config, *args, **kwargs)


def _instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
    """
    :param config: An config object describing what to call and what params to use.
                   In addition to the parameters, the config must contain:
                   _target_ : target class or callable name (str)
                   And may contain:
                   _args_: List-like of positional arguments to pass to the target
                   _recursive_: Construct nested objects as well (bool).
                                False by default.
                                may be overridden via a _recursive_ key in
                                the kwargs
                   _convert_: Conversion strategy
                        none    : Passed objects are DictConfig and ListConfig, default
                        partial : Passed objects are converted to dict and list, with
                                  the exception of Structured Configs (and their fields).
                        all     : Passed objects are dicts, lists and primitives without
                                  a trace of OmegaConf containers
                   _partial_: If True, return functools.partial wrapped method or object
                              False by default. Configure per target.
                   _args_: List-like of positional arguments
    :param args: Optional positional parameters pass-through
    :param kwargs: Optional named parameters to override
                   parameters in the config object. Parameters not present
                   in the config objects are being passed as is to the target.
                   IMPORTANT: dataclasses instances in kwargs are interpreted as config
                              and cannot be used as passthrough
    :return: if _target_ is a class name: the instantiated object
             if _target_ is a callable: the return value of the call
    """
    verbose = config.get(SpecialKeys.VERBOSE, False)
    if not __global_config__.__initilized__:
        __global_config__.initialize()
    if not _is_instantiatable(config):
        if verbose:
            logger.info("Config is not instantiatable, returning config")
        return config
    _recursive_ = config.get(SpecialKeys.RECURSIVE, False)
    if SpecialKeys.RECURSIVE not in kwargs:
        kwargs[SpecialKeys.RECURSIVE.value] = _recursive_
    if verbose:
        logger.info("instantiating %s ...", config.get(SpecialKeys.TARGET))
    return hydra.utils.instantiate(config, *args, **kwargs)


def _methods(cfg: Any, obj: object, return_function=False):
    cfg = _to_dict(cfg)
    if not cfg:
        logger.info("No method defined to call")
        return

    if isinstance(cfg, dict) and SpecialKeys.METHOD in cfg:
        _method_ = cfg[SpecialKeys.METHOD]
    elif isinstance(cfg, dict):
        _method_ = cfg
    elif isinstance(cfg, str):
        _method_ = cfg
        cfg = {}
    else:
        raise ValueError(f"Invalid method: {cfg}")

    if isinstance(_method_, str):
        _fn = getattr(obj, _method_)
        if return_function:
            logger.info(f"Returning function {_fn}")
            return _fn
        logger.info(f"Calling {_method_}")
        return _fn(**cfg)
    elif isinstance(_method_, dict):
        if SpecialKeys.CALL in _method_:
            _call_ = _method_.pop(SpecialKeys.CALL)
        else:
            _call_ = True
        if _call_:
            _fn = getattr(obj, _method_[SpecialKeys.METHOD_NAME])
            _parms = _method_.pop(SpecialKeys.rcPARAMS, {})
            if return_function:
                if not _parms:
                    logger.info(f"Returning function {_fn}")
                    return _fn
                logger.info(f"Returning function {_fn} with params {_parms}")
                return functools.partial(_fn, **_parms)
            logger.info(f"Calling {_method_}")
            return _fn(**_parms)
        else:
            logger.info(f"Skipping call to {_method_}")
    elif isinstance(_method_, list):
        for _each_method in _method_:
            logger.info(f"Calling {_each_method}")
            if isinstance(_each_method, str):
                getattr(obj, _each_method)()
            elif isinstance(_each_method, dict):
                if SpecialKeys.CALL in _each_method:
                    _call_ = _each_method.pop(SpecialKeys.CALL)
                else:
                    _call_ = True
                if _call_:
                    getattr(obj, _each_method[SpecialKeys.METHOD_NAME])(
                        **_each_method[SpecialKeys.rcPARAMS]
                    )
                else:
                    logger.info(f"Skipping call to {_each_method}")


def _function(cfg: Any, _name_, return_function=False, **parms):
    cfg = _to_dict(cfg)
    if not isinstance(cfg, dict):
        logger.info("No function defined to execute")
        return None

    if SpecialKeys.FUNC not in cfg:
        logger.info("No function defined to execute")
        return None

    _functions_ = cfg[SpecialKeys.FUNC]
    fn = _partial(_functions_[_name_])
    if _name_ in cfg:
        _parms = cfg[_name_]
        _parms = {**_parms, **parms}
    else:
        _parms = parms
    _exec_ = _parms.pop(SpecialKeys.EXEC) if SpecialKeys.EXEC in _parms else True
    if _exec_:
        if callable(fn):
            if return_function:
                logger.info(f"Returning function {fn}")
                return fn
            logger.info(f"Executing function {fn} with parms {_parms}")
            return fn(**_parms)
        else:
            logger.info(f"Function {_name_} not callable")
            return None
    else:
        logger.info(f"Skipping execute of {fn}")
        return None


def __search_package_path__():
    return __global_config__.hyfi_config_path


OmegaConf.register_new_resolver("__hyfi_path__", __hyfi_path__)
OmegaConf.register_new_resolver("__search_package_path__", __search_package_path__)
OmegaConf.register_new_resolver("__home_path__", __home_path__)
OmegaConf.register_new_resolver("today", today)
OmegaConf.register_new_resolver("to_datetime", strptime)
OmegaConf.register_new_resolver("iif", lambda cond, t, f: t if cond else f)
OmegaConf.register_new_resolver("alt", lambda val, alt: val or alt)
OmegaConf.register_new_resolver("randint", random.randint, use_cache=True)
OmegaConf.register_new_resolver("get_method", hydra.utils.get_method)
OmegaConf.register_new_resolver("get_original_cwd", getcwd)
OmegaConf.register_new_resolver("exists", exists)
OmegaConf.register_new_resolver("join_path", join_path)
OmegaConf.register_new_resolver("mkdir", mkdir)
OmegaConf.register_new_resolver("dirname", os.path.dirname)
OmegaConf.register_new_resolver("basename", os.path.basename)
OmegaConf.register_new_resolver("check_path", check_path)
OmegaConf.register_new_resolver("cached_path", cached_path)
OmegaConf.register_new_resolver(
    "lower_case_with_underscores", lower_case_with_underscores
)
OmegaConf.register_new_resolver("dotenv_values", dotenv_values)


def _getsource(obj):
    """Return the source code of the object."""
    try:
        if _is_config(obj):
            if SpecialKeys.TARGET in obj:
                target_string = obj[SpecialKeys.TARGET]
                mod_name, object_name = target_string.rsplit(".", 1)
                mod = importlib.import_module(mod_name)
                obj = getattr(mod, object_name)
        elif isinstance(obj, str):
            mod_name, object_name = obj.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            obj = getattr(mod, object_name)
        return inspect.getsource(obj)
    except Exception as e:
        logger.error(f"Error getting source: {e}")
        return ""


def _viewsource(obj):
    """Print the source code of the object."""
    print(_getsource(obj))


def _ensure_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    return _to_dict(value)


def _ensure_kwargs(_kwargs, _fn):
    from inspect import getfullargspec as getargspec

    if callable(_fn):
        args = getargspec(_fn).args
        logger.info(f"args of {_fn}: {args}")
        return {k: v for k, v in _kwargs.items() if k in args}
    return _kwargs



"""Motion image processing functions."""
import os
import subprocess
from pathlib import Path

from hyfi.utils.file import get_filepaths
from hyfi.utils.logging import getLogger
from hyfi.utils.notebook import display_image

logger = getLogger(__name__)


def make_gif(
    image_filepaths=None,
    filename_patterns=None,
    base_dir=None,
    output_filepath=None,
    duration=100,
    loop=0,
    width=None,
    optimize=True,
    quality=50,
    show=False,
    force=False,
    **kwargs,
):
    """
    Create a GIF from a list of images or a list of filenames.
    """
    from PIL import Image

    logger.info(f"Making GIF from {filename_patterns}")
    if os.path.exists(output_filepath) and not force:
        logger.info(f"Skipping GIF creation, already exists: {output_filepath}")
        logger.info("If you want to re-create the GIF, set force=True")
    else:
        if image_filepaths is None:
            image_filepaths = sorted(
                get_filepaths(filename_patterns, base_dir=base_dir)
            )
        if not image_filepaths:
            logger.warning("no images found")
            return
        if frames := [Image.open(image) for image in image_filepaths]:
            frame_one = frames[0]
            frame_one.save(
                output_filepath,
                format="GIF",
                append_images=frames,
                save_all=True,
                duration=duration,
                loop=loop,
                optimize=optimize,
                quality=quality,
            )
            print(f"Saved GIF to {output_filepath}")
        else:
            logger.warning(f"No frames found for {filename_patterns}")

    if show and os.path.exists(output_filepath):
        display_image(data=open(output_filepath, "rb").read(), width=width)

    return output_filepath


def extract_frames(
    video_path, extract_nth_frame, extracted_frame_dir, frame_filename="%04d.jpg"
):
    """
    Extract frames from a video.
    """
    logger.info(f"Exporting Video Frames (1 every {extract_nth_frame})...")
    try:
        for f in Path(f"{extracted_frame_dir}").glob("*.jpg"):
            f.unlink()
    except FileNotFoundError:
        logger.info(f"No video frames found in {extracted_frame_dir}")
    vf = f"select=not(mod(n\,{extract_nth_frame}))"

    ffmpeg_path = "/usr/bin/ffmpeg"
    if not os.path.exists(ffmpeg_path):
        ffmpeg_path = "ffmpeg"
    if os.path.exists(video_path):
        subprocess.run(
            [
                ffmpeg_path,
                "-i",
                f"{video_path}",
                "-vf",
                f"{vf}",
                "-vsync",
                "vfr",
                "-q:v",
                "2",
                "-loglevel",
                "error",
                "-stats",
                f"{extracted_frame_dir}/{frame_filename}",
            ],
            stdout=subprocess.PIPE,
        ).stdout.decode("utf-8")
    else:
        logger.warning(
            f"WARNING!\n\nVideo not found: {video_path}.\nPlease check your video path."
        )


def create_video(
    base_dir, video_path, input_url, fps, start_number, vframes, force=False
):
    """
    Create a video from a list of images.
    """

    logger.info(f"Creating video from {input_url}")
    if os.path.exists(video_path) and not force:
        logger.info(f"Skipping video creation, already exists: {video_path}")
        logger.info("If you want to re-create the video, set force=True")
        return video_path

    cmd = [
        "ffmpeg",
        "-y",
        "-vcodec",
        "png",
        "-r",
        str(fps),
        "-start_number",
        str(start_number),
        "-i",
        input_url,
        "-frames:v",
        str(vframes),
        "-c:v",
        "libx264",
        "-vf",
        f"fps={fps}",
        "-pix_fmt",
        "yuv420p",
        "-crf",
        "17",
        "-preset",
        "veryslow",
        video_path,
    ]

    process = subprocess.Popen(
        cmd,
        cwd=f"{base_dir}",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = process.communicate()
    if process.returncode != 0:
        print(stderr)
        raise RuntimeError(stderr)
    else:
        print(f"The video is ready and saved to {video_path}")

    return video_path



from omegaconf import DictConfig

from hyfi.batch import BatchConfig
from hyfi.hydra import _to_dict
from hyfi.hydra.main import _load, _merge, _print, _save, _save_json
from hyfi.tasks.base import BaseTask
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class BatchTask(BaseTask):
    batch: BatchConfig = None  # type: ignore

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
        validate_assignment = False
        exclude = {
            "_config_",
            "_initial_config_",
            "__data__",
            "path",
            "module",
            "secret",
            "auto",
            "project",
        }
        include = {}
        underscore_attrs_are_private = True
        property_set_methods = {
            "name": "set_name",
            "batch_name": "set_name",
            "batch_num": "set_batch_num",
            "root_dir": "set_root_dir",
        }

    def __init__(self, **args):
        super().__init__(**args)

    def set_name(self, val):
        super().set_name(val)
        self._config_.batch.batch_name = val
        self.batch.batch_name = val
        self.initialize_configs(name=val)

    def set_batch_num(self, val):
        self._config_.batch.batch_num = val
        self.batch.batch_num = val

    def initialize_configs(
        self, root_dir=None, batch_config_class=BatchConfig, **kwargs
    ):
        super().initialize_configs(root_dir=root_dir, **kwargs)

        self.batch = batch_config_class(**self.config.batch)
        self.batch_num = self.batch.batch_num
        # if self.project.use_huggingface_hub:
        #     self.secrets.init_huggingface_hub()
        if self.verbose:
            logger.info(
                f"Initalized batch: {self.batch_name}({self.batch_num}) in {self.root_dir}"
            )

    @property
    def batch_name(self):
        return self.batch.batch_name

    @property
    def batch_num(self):
        return self.batch.batch_num

    @property
    def seed(self):
        return self.batch.seed

    @property
    def batch_dir(self):
        return self.batch.batch_dir

    @property
    def dataset_dir(self):
        return self.path.dataset_dir

    @property
    def verbose(self):
        return self.batch.verbose

    @property
    def device(self):
        return self.batch.device

    @property
    def num_devices(self):
        return self.batch.num_devices

    def save_config(
        self,
        config=None,
        exclude=None,
        include=None,
    ):
        """Save the batch config"""
        if config is not None:
            self._config_ = config
        logger.info(f"Saving config to {self.batch.config_filepath}")
        cfg = _to_dict(self.config)
        if exclude is None:
            exclude = self.__config__.exclude

        if include:
            if isinstance(include, str):
                include = [include]
            args = {key: cfg[key] for key in include}
        else:
            args = cfg
            if exclude:
                if isinstance(exclude, str):
                    exclude = [exclude]
                for key in exclude:
                    args.pop(key, None)
        _save(args, self.batch.config_filepath)
        self.save_settings(exclude=exclude)
        return self.batch.config_filename

    def save_settings(self, exclude=None, exclude_none=True):
        def dumper(obj):
            return _to_dict(obj) if isinstance(obj, DictConfig) else str(obj)

        if exclude is None:
            exclude = self.__config__.exclude
        config = self.dict(exclude=exclude, exclude_none=exclude_none)
        if self.verbose:
            logger.info(f"Saving config to {self.batch.config_jsonpath}")
        _save_json(config, self.batch.config_jsonpath, default=dumper)

    def load_config(
        self,
        batch_name=None,
        batch_num=None,
        **args,
    ):
        """Load the config from the batch config file"""
        if self.verbose:
            logger.info(
                f"> Loading config for batch_name: {batch_name} batch_num: {batch_num}"
            )
        # self.config.batch.batch_num = batch_num
        if batch_name is None:
            batch_name = self.batch_name

        if batch_num is not None:
            cfg = self._initial_config_.copy()
            self.batch.batch_name = batch_name
            self.batch.batch_num = batch_num
            _path = self.batch.config_filepath
            if _path.is_file():
                logger.info(f"Loading config from {_path}")
                batch_cfg = _load(_path)
                if self.verbose:
                    logger.info("Merging config with the loaded config")
                cfg = _merge(cfg, batch_cfg)
            else:
                logger.info(f"No config file found at {_path}")
                batch_num = None
        else:
            cfg = self.config

        if self.verbose:
            logger.info(f"Merging config with args: {args}")
        self._config_ = _merge(cfg, args)

        self.batch_num = batch_num
        self.batch_name = batch_name

        return self.config

    def show_config(self, batch_name=None, batch_num=None):
        self.load_config(batch_name, batch_num)
        _print(self.dict())



"""Utilities for working with notebooks.""" ""
import logging
import os
import sys

from hyfi.utils.file import read

logger = logging.getLogger(__name__)


def is_notebook():
    """Check if the code is running in a notebook."""
    try:
        get_ipython
    except NameError:
        return False
    # pylint: disable=undefined-variable
    shell_type = get_ipython().__class__.__name__  # noqa
    # logger.info(f"shell type: {shell_type}")
    return shell_type in ["ZMQInteractiveShell", "Shell"]


def is_colab():
    """Check if the code is running in Google Colab."""
    is_colab = "google.colab" in sys.modules
    if is_colab:
        logger.info("Google Colab detected.")
    else:
        logger.info("Google Colab not detected.")
    return is_colab


def get_display():
    """Get the display object for the current environment."""
    try:
        from ipywidgets import Output
    except ImportError:
        logger.info("ipywidgets not installed.")
        return None

    return Output() if is_notebook() else None


def clear_output(wait=False):
    """Clear the output of the current notebook."""
    from IPython import display

    if is_notebook():
        display.clear_output(wait=wait)


def display(
    *objs,
    include=None,
    exclude=None,
    metadata=None,
    transient=None,
    display_id=None,
    **kwargs,
):
    """Display an object in the current notebook."""
    from IPython import display

    if is_notebook() and objs is not None:
        return display.display(
            *objs,
            include=include,
            exclude=exclude,
            metadata=metadata,
            transient=transient,
            display_id=display_id,
            **kwargs,
        )


def display_image(
    data=None,
    url=None,
    filename=None,
    format=None,
    embed=None,
    width=None,
    height=None,
    retina=False,
    unconfined=False,
    metadata=None,
    **kwargs,
):
    """
    Display an image, which can be given as raw data or a URL.

    Parameters
    ----------
    data : unicode, str or bytes
        The raw image data or a URL or filename to load the data from.
        This always results in embedded image data.
    url : unicode
        A URL to download the data from. If you specify `url=`,
        the image data will not be embedded unless you also specify `embed=True`.
    filename : unicode
        Path to a local file to load the data from.
        Images from a file are always embedded.
    format : unicode
        The format of the image data (png/jpeg/jpg/gif). If a filename or URL is given
        for format will be inferred from the filename extension.
    embed : bool
        Should the image data be embedded using a data URI (True) or be
        loaded using an <img> tag. Set this to True if you want the image
        to be viewable later with no internet connection in the notebook.

        Default is `True`, unless the keyword argument `url` is set, then
        default value is `False`.

        Note that QtConsole is not able to display images if `embed` is set to `False`
    width : int
        Width in pixels to which to constrain the image in html
    height : int
        Height in pixels to which to constrain the image in html
    retina : bool
        Automatically set the width and height to half of the measured
        width and height.
        This only works for embedded images because it reads the width/height
        from image data.
        For non-embedded images, you can just set the desired display width
        and height directly.
    unconfined: bool
        Set unconfined=True to disable max-width confinement of the image.
    metadata: dict
        Specify extra metadata to attach to the image.

    """
    from IPython import display

    if is_notebook():
        img = display.Image(
            data=data,
            url=url,
            filename=filename,
            format=format,
            embed=embed,
            width=width,
            height=height,
            retina=retina,
            unconfined=unconfined,
            metadata=metadata,
            **kwargs,
        )
        return display.display(img)


def hide_code_in_slideshow():
    """Hide code in slideshow."""
    import binascii

    from IPython import display

    uid = binascii.hexlify(os.urandom(8)).decode()
    html = """<div id="%s"></div>
    <script type="text/javascript">
        $(function(){
            var p = $("#%s");
            if (p.length==0) return;
            while (!p.hasClass("cell")) {
                p=p.parent();
                if (p.prop("tagName") =="body") return;
            }
            var cell = p;
            cell.find(".input").addClass("hide-in-slideshow")
        });
    </script>""" % (
        uid,
        uid,
    )
    display.display_html(html, raw=True)


def colored_str(s, color="black"):
    """Colored string."""
    # return "<text style=color:{}>{}</text>".format(color, s)
    return "<text style=color:{}>{}</text>".format(color, s.replace("\n", "<br>"))


def cprint(str_tuples):
    from IPython.display import HTML as html_print
    from IPython.display import display

    display(html_print(" ".join([colored_str(ti, color=ci) for ti, ci in str_tuples])))


def create_dropdown(
    options, value, description, disabled=False, style=None, layout=None, **kwargs
):
    """Create a dropdown widget."""
    if style is None:
        style = {"description_width": "initial"}
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    return widgets.Dropdown(
        options=options,
        value=value,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )


def create_textarea(
    value,
    description,
    placeholder="",
    disabled=False,
    style=None,
    layout=None,
    **kwargs,
):
    """Create a textarea widget."""
    if style is None:
        style = {"description_width": "initial"}
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    return widgets.Textarea(
        value=value,
        placeholder=placeholder,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )


def create_button(description, button_style="", icon="check", layout=None, **kwargs):
    """Create a button widget."""
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    return widgets.Button(
        description=description,
        button_style=button_style,
        icon=icon,
        layout=layout,
        **kwargs,
    )


def create_radiobutton(
    options, description, value=None, disabled=False, style=None, layout=None, **kwargs
):
    """Create a radiobutton widget."""
    if style is None:
        style = {"description_width": "initial"}
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    return widgets.RadioButtons(
        options=options,
        value=value,
        description=description,
        disabled=disabled,
        style=style,
        layout=layout,
        **kwargs,
    )


def create_image(
    filename=None,
    format=None,
    width=None,
    height=None,
    **kwargs,
):
    """Create an image widget."""
    import ipywidgets as widgets

    # from urllib.request import urlopen

    if filename is None:
        url = "https://assets.entelecheia.cc/img/placeholder.png"
        # img = urlopen(url).read()
        img = read(url)
        _format = "png"
    else:
        img = read(filename)
        _format = format or filename.split(".")[-1]
    return widgets.Image(
        value=img,
        format=_format,
        width=width,
        height=height,
        **kwargs,
    )


def create_floatslider(
    min=0.0,
    max=1.0,
    step=0.1,
    value=None,
    description="",
    disabled=False,
    continuous_update=False,
    orientation="horizontal",
    readout=True,
    readout_format=".1f",
    style=None,
    layout=None,
    **kwargs,
):
    """Create a float slider widget."""
    if style is None:
        style = {"description_width": "initial"}
    import ipywidgets as widgets

    layout = (
        widgets.Layout(width="auto") if layout is None else widgets.Layout(**layout)
    )
    return widgets.FloatSlider(
        min=min,
        max=max,
        step=step,
        value=value,
        description=description,
        disabled=disabled,
        continuous_update=continuous_update,
        orientation=orientation,
        readout=readout,
        readout_format=readout_format,
        style=style,
        layout=layout,
        **kwargs,
    )


def load_extentions(exts=None):
    """Load extentions."""
    if exts is None:
        exts = ["autotime"]
    if not is_notebook():
        return
    from IPython import get_ipython

    ip = get_ipython()
    try:
        loaded = ip.extension_manager.loaded
        for ext in exts:
            if ext not in loaded:
                ip.extentension_manager.load_extension(ext)
    except AttributeError:
        for ext in exts:
            try:
                ip.magic(f"load_ext {ext}")
            except ModuleNotFoundError:
                logger.info("Extension %s not found. Install it first.", ext)


def set_matplotlib_formats(*formats, **kwargs):
    """Set matplotlib formats."""
    if is_notebook():
        from IPython.display import set_matplotlib_formats

        set_matplotlib_formats(*formats, **kwargs)



from pathlib import Path

from pydantic import BaseModel

from hyfi.about import AboutConfig
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)

__hydra_version_base__ = "1.2"

__about__ = AboutConfig()


class HydraConfig(BaseModel):
    """HyFI config primary class"""

    hyfi_config_path: str = __about__.config_path
    hyfi_config_module: str = __about__.config_module
    hyfi_user_config_path: str = ""


__hydra_config__ = HydraConfig()


def __hyfi_path__():
    return Path(__file__).parent.parent.as_posix()


def __home_path__():
    return Path.home().as_posix()



"""Plotting functions"""
import os
import platform
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import font_manager, rc

from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def get_plot_font(
    set_font_for_matplot=True, fontpath=None, fontname=None, verbose=False
):
    """Get font for plot"""
    if fontname and not fontname.endswith(".ttf"):
        fontname += ".ttf"
    if not fontpath:
        if platform.system() == "Darwin":
            fontname = fontname or "AppleGothic.ttf"
            fontpath = os.path.join("/System/Library/Fonts/Supplemental/", fontname)
        elif platform.system() == "Windows":
            fontname = fontname or "malgun.ttf"
            fontpath = os.path.join("c:/Windows/Fonts/", fontname)
        elif platform.system() == "Linux":
            fontname = fontname or "NanumGothic.ttf"
            if fontname.lower().startswith("nanum"):
                fontpath = os.path.join("/usr/share/fonts/truetype/nanum/", fontname)
            else:
                fontpath = os.path.join("/usr/share/fonts/truetype/", fontname)
        if fontpath and not Path(fontpath).is_file():
            paths = find_font_file(fontname)
            fontpath = paths[0] if len(paths) > 0 else None
        if verbose:
            logger.info(f"Font path: {fontpath}")

    if fontpath is None or not Path(fontpath).is_file():
        fontname = None
        fontpath = None
        logger.warning(f"Font file does not exist at {fontpath}")
        if platform.system() == "Linux":
            font_install_help = """
            apt install fontconfig
            apt install fonts-nanum
            fc-list | grep -i nanum
            """
            print(font_install_help)
    else:
        font_manager.fontManager.addfont(fontpath)
        fontname = font_manager.FontProperties(fname=fontpath).get_name()

        if set_font_for_matplot and fontname:
            rc("font", family=fontname)
            plt.rcParams["axes.unicode_minus"] = False
            if verbose:
                font_family = plt.rcParams["font.family"]
                logger.info(f"font family: {font_family}")
        if verbose:
            logger.info(f"font name: {fontname}")
    return fontname, fontpath


def find_font_file(query):
    """Find font file by query string"""
    return list(
        filter(
            lambda path: query in os.path.basename(path),
            font_manager.findSystemFonts(),
        )
    )



from pathlib import Path
from typing import Any

from pydantic import BaseModel

from hyfi.__global__ import __about__
from hyfi.hydra import _compose
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class PathConfig(BaseModel):
    config_name: str = "__init__"
    # internal paths for hyfi
    home: str = ""
    hyfi: str = ""
    resources: str = ""
    runtime: str = ""
    # global paths
    global_hyfi_root: str = ""
    global_workspace_name: str = "workspace"
    global_workspace_root: str = ""
    global_archive: str = ""
    global_datasets: str = ""
    global_models: str = ""
    global_modules: str = ""
    global_library: str = ""
    global_cache: str = ""
    global_tmp: str = ""
    # project specific paths
    project_root: str = ""
    project_workspace_name: str = "workspace"
    project_workspace_root: str = ""
    project_archive: str = ""
    project_datasets: str = ""
    project_models: str = ""
    project_modules: str = ""
    project_outputs: str = ""
    project_logs: str = ""
    project_library: str = ""
    project_cache: str = ""
    project_tmp: str = ""

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        """
        Initialize the config. This is the base implementation of __init__. You can override this in your own subclass if you want to customize the initilization of a config by passing a keyword argument ` data `.

        Args:
                config_name: The name of the config to initialize
                data: The data to initialize
        """
        # Initialize the config module.
        data = _compose(
            f"path={config_name}",
            config_data=data,
            config_module=__about__.config_module,
        )  # type: ignore
        super().__init__(config_name=config_name, **data)

    @property
    def log_dir(self):
        """
        Create and return the path to the log directory. This is a convenience method for use in unit tests that want to ensure that the log directory exists and is accessible to the user.


        Returns:
                absolute path to the log directory for the project ( including parent directories
        """
        Path(self.project_logs).mkdir(parents=True, exist_ok=True)
        return Path(self.project_logs).absolute()

    @property
    def cache_dir(self):
        """
        Create and return the directory where cache files are stored. This is useful for debugging and to ensure that we don't accidentally delete the cache files when there are too many files in the cache.


        Returns:
                absolute path to the cache directory for this test run
        """
        Path(self.global_cache).mkdir(parents=True, exist_ok=True)
        return Path(self.global_cache).absolute()



"""Environment variable utilities"""
import os
from collections import defaultdict
from pathlib import Path
from string import Template
from typing import Any, Union

import dotenv
import hydra

from hyfi.utils.file import is_dir
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def getcwd():
    """Get the original working directory before Hydra changed it.

    This function tries to call the `get_original_cwd` function from the `hydra.utils` module,
    which returns the original working directory if it exists. If the `get_original_cwd` function
    raises a `ValueError` exception, it means that Hydra did not change the working directory,
    so the function falls back to calling the `os.getcwd` function, which returns the current
    working directory.

    Returns:
        str: The original working directory before Hydra changed it.
    """
    try:
        return hydra.utils.get_original_cwd()
    except ValueError:
        return os.getcwd()


def expand_posix_vars(posix_expr: str, context: dict = None) -> str:  # type: ignore
    # sourcery skip: dict-assign-update-to-union
    """
    Expand POSIX variables in a string.

    Args:
        posix_expr (str): The string containing POSIX variables to be expanded.
        context (dict, optional): A dictionary containing additional variables to be used in the expansion.
            Defaults to None.

    Returns:
        str: The expanded string.

    Examples:
        >>> expand_posix_vars("$HOME")
        '/home/user'
        >>> expand_posix_vars("$HOME/$USER", {"USER": "testuser"})
        '/home/user/testuser'

    """
    # Set the context to the default context.
    if context is None:
        context = {}
    env = defaultdict(str, os.environ.copy())
    env.update(context)
    return Template(posix_expr).substitute(env)


def dotenv_values(dotenv_path: str = "", **kwargs):
    """
    Load dotenv file and return a dict of key / value pairs. This is a wrapper around : py : func : ` dotenv. dotenv_values `

    Args:
        dotenv_path: path to. env file

    Returns:
        dict of key / value pairs ( key = value )
    """
    config = dotenv.dotenv_values(dotenv_path=dotenv_path, **kwargs)
    return dict(config)


def load_dotenv(
    override: bool = False,
    dotenv_dir: str = "",
    dotenv_filename: str = ".env",
    verbose: bool = False,
) -> None:
    """
    Load. env file from given directory or from current directory. This is a convenience function for use in tests that want to run dotenv in a non - interactive environment

    Args:
        override: If True override existing. env file
        dotenv_dir: Directory to look for. env file ( default cwd )
        dotenv_filename: Name of. env file to look for
        verbose: Print debug information to console

    Returns:
        None or a Path object for the. env file
    """
    dotenv_dir = dotenv_dir or getcwd()
    dotenv_path = Path(dotenv_dir, dotenv_filename)
    # Load. env files and directories.
    if dotenv_path.is_file():
        dotenv.load_dotenv(dotenv_path=dotenv_path, verbose=verbose, override=override)
        os.environ["DOTENV_PATH"] = str(dotenv_path)
        os.environ["DOTENV_DIR"] = str(dotenv_path.parent)
        # Load. env from dotenv_path.
        if verbose:
            logger.info("Loaded .env from %s", dotenv_path)
        else:
            logger.debug("Loaded .env from %s", dotenv_path)
    else:
        # If verbose is true print out the. env file.
        if verbose:
            logger.info(
                "No .env file found in %s, finding .env in parent dirs", dotenv_path
            )
        else:
            logger.debug(
                "No .env file found in %s, finding .env in parent dirs", dotenv_path
            )
        # Load dotenv. env from dotenv.
        if dotenv_path := dotenv.find_dotenv():
            dotenv.load_dotenv(
                dotenv_path=dotenv_path, verbose=verbose, override=override
            )
            os.environ["DOTENV_PATH"] = str(dotenv_path)
            os.environ["DOTENV_DIR"] = os.path.dirname(dotenv_path)
            # Load. env from dotenv_path.
            if verbose:
                logger.info("Loaded .env from %s", dotenv_path)
            else:
                logger.debug("Loaded .env from %s", dotenv_path)
        else:
            os.environ["DOTENV_PATH"] = ""
            os.environ["DOTENV_DIR"] = ""
            # Print out the. env file if verbose is true.
            if verbose:
                logger.info("No .env file found in %s", dotenv_path)
            else:
                logger.debug("No .env file found in %s", dotenv_path)


def get_osenv(key: str = "", default: Union[str, None] = None) -> Any:
    """Get the value of an environment variable or return the default value"""
    load_dotenv()
    return os.environ.get(key, default) if key else os.environ


def set_osenv(key: str, value: Any) -> None:
    """Set the value of an environment variable"""
    if value and is_dir(value):
        value = os.path.abspath(value)
    if pre_val := os.environ.get(key):
        logger.info("Overwriting %s=%s with %s", key, pre_val, value)
    else:
        logger.info("Setting %s=%s", key, value)
    os.environ[key] = value


def _check_and_set_value(key: str, value: Any) -> Any:
    """Check and set value to environment variable"""
    env_key = key.upper()
    if value is not None:
        old_value = os.getenv(env_key, "")
        if str(old_value).lower() != str(value).lower():
            os.environ[env_key] = str(value)
            logger.debug("Set environment variable %s=%s", env_key, str(value))
    return value



from pathlib import Path
from typing import Any

from pydantic import BaseModel

from hyfi.hydra import _compose
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class BatchPathConfig(BaseModel):
    config_name: str = "__batch__"

    project_workspace_root = ""
    batch_name: str = "demo"
    batch_output: str = ""
    task_name: str = "default-task"
    task_root: str = ""
    task_outputs: str = ""
    task_datasets: str = ""
    task_library: str = ""
    task_models: str = ""
    task_cache: str = ""
    task_tmp: str = ""
    task_log: str = ""

    class Config:
        extra = "ignore"

    def __init__(self, config_name: str = "__batch__", **data: Any):
        """
        Initialize the batch. This is the method you call when you want to initialize the batch from a config

        Args:
                config_name: The name of the config you want to use
                data: The data you want to initilize the
        """
        # Initialize the config with the given config_name.
        if not data:
            logger.debug(
                "There are no arguments to initilize a config, using default config %s",
                config_name,
            )
            data = _compose(f"path={config_name}")
        super().__init__(**data)

    @property
    def project_workspace_dir(self) -> Path:
        """
        Get the path to the project workspace directory.


        Returns:
                absolute path to the project workspace directory or None if not set by the user ( in which case a default is used )
        """
        self.project_workspace_root = self.project_workspace_root or "./workspace"
        return Path(self.project_workspace_root).absolute()

    @property
    def root_dir(self) -> Path:
        """
        Returns the absolute path to the task root directory. If the task_root attribute is set it is used as the path to the task's data directory. Otherwise the task name is used as the path to the project's data directory.


        Returns:
                an absolute path to the task root directory or None if it doesn't exist or cannot be converted to a path object
        """
        self.task_root = (
            self.task_root or (self.project_workspace_dir / self.task_name).as_posix()
        )
        # return as an absolute path
        return Path(self.task_root).absolute()

    @property
    def output_dir(self) -> Path:
        """
        Returns the directory where the task outputs are stored. It is used to determine where the output files will be stored when running the task.


        Returns:
                absolute path to the output directory of the task ( relative to the task root directory ) or None if not
        """
        self.task_outputs = self.task_outputs or (self.root_dir / "outputs").as_posix()
        return Path(self.task_outputs).absolute()

    @property
    def batch_dir(self) -> Path:
        """
        The directory where the batch is stored. It is used to determine where the results are stored for a batch of data to be processed.


        Returns:
                The directory where the batch output is stored for a batch of data to be processed ( relative to the task output directory )
        """
        self.batch_output = (
            self.batch_output or (self.output_dir / self.batch_name).as_posix()
        )
        return Path(self.batch_output).absolute()

    @property
    def library_dir(self) -> Path:
        """
        The path to the library.


        Returns:
                absolute path to the library directory ( relative to the task root directory ).
        """
        self.task_library = self.task_library or (self.root_dir / "library").as_posix()
        return Path(self.task_library).absolute()

    @property
    def dataset_dir(self) -> Path:
        """
        Get the path to the dataset directory.


        Returns:
                absolute path to the dataset directory ( relative to the root directory ) or None if not set in the
        """
        self.task_datasets = (
            self.task_datasets or (self.root_dir / "datasets").as_posix()
        )
        return Path(self.task_datasets).absolute()

    @property
    def model_dir(self) -> Path:
        """
        Get the directory where models are stored.


        Returns:
                Absolute path to the models directory on the task's filesystem ( relative to the root_dir )
        """
        self.task_models = self.task_models or (self.root_dir / "models").as_posix()
        return Path(self.task_models).absolute()

    @property
    def cache_dir(self) -> Path:
        """
        The directory where tasks are cached.


        Returns:
                A path to the cache directory for this task or None if it is not set in the config
        """
        self.task_cache = self.task_cache or (self.root_dir / "cache").as_posix()
        return Path(self.task_cache).absolute()

    @property
    def tmp_dir(self) -> Path:
        """
        Returns the path to the temporary directory.


        Returns:
                absolute path to the temporary directory of the task ( relative to the root_dir ).
        """
        self.task_tmp = self.task_tmp or (self.root_dir / "tmp").as_posix()
        return Path(self.task_tmp).absolute()

    @property
    def log_dir(self) -> Path:
        """
        Get the path to the log directory.


        Returns:
                absolute path to the log directory of the task ( relative to the root_dir ).
        """
        self.task_log = self.task_log or (self.root_dir / "logs").as_posix()
        log_dir = Path(self.task_log).absolute()
        log_dir.mkdir(parents=True, exist_ok=True)
        return log_dir



import os
from typing import Optional, Tuple, Union

from pydantic import BaseSettings, SecretStr, root_validator
from pydantic.env_settings import SettingsSourceCallable

from hyfi.utils.env import load_dotenv
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class DotEnvConfig(BaseSettings):
    """Environment variables for HyFI"""

    config_name: str = "__init__"

    DOTENV_FILENAME: Optional[str] = ".env"
    DOTENV_DIR: Optional[str] = ""
    DOTENV_PATH: Optional[str] = ""
    # Internal
    HYFI_RESOURCE_DIR: Optional[str] = ""
    HYFI_GLOBAL_ROOT: Optional[str] = ""
    HYFI_GLOBAL_WORKSPACE_NAME: Optional[str] = "workspace"
    HYFI_PROJECT_NAME: Optional[str] = ""
    HYFI_TASK_NAME: Optional[str] = ""
    HYFI_PROJECT_DESC: Optional[str] = ""
    HYFI_PROJECT_ROOT: Optional[str] = ""
    HYFI_PROJECT_WORKSPACE_NAME: Optional[str] = "workspace"
    HYFI_LOG_LEVEL: Optional[str] = "WARNING"
    HYFI_VERBOSE: Optional[Union[bool, str, int]] = False
    HYFI_NUM_WORKERS: Optional[int] = 1
    CACHED_PATH_CACHE_ROOT: Optional[str] = ""
    # For other packages
    CUDA_DEVICE_ORDER: Optional[str] = "PCI_BUS_ID"
    CUDA_VISIBLE_DEVICES: Optional[str] = ""
    WANDB_PROJECT: Optional[str] = ""
    WANDB_DISABLED: Optional[str] = ""
    WANDB_DIR: Optional[str] = ""
    WANDB_NOTEBOOK_NAME: Optional[str] = ""
    WANDB_SILENT: Optional[Union[bool, str]] = False
    LABEL_STUDIO_SERVER: Optional[str] = ""
    KMP_DUPLICATE_LIB_OK: Optional[str] = "True"
    TOKENIZERS_PARALLELISM: Optional[Union[bool, str]] = False
    # API Keys and Tokens
    WANDB_API_KEY: Optional[SecretStr] = None
    HUGGING_FACE_HUB_TOKEN: Optional[SecretStr] = None
    OPENAI_API_KEY: Optional[SecretStr] = None
    ECOS_API_KEY: Optional[SecretStr] = None
    FRED_API_KEY: Optional[SecretStr] = None
    NASDAQ_API_KEY: Optional[SecretStr] = None
    HF_USER_ACCESS_TOKEN: Optional[SecretStr] = None
    LABEL_STUDIO_USER_TOKEN: Optional[SecretStr] = None

    class Config:
        env_prefix = ""
        env_nested_delimiter = "__"
        case_sentive = False
        env_file = ".env"
        env_file_encoding = "utf-8"
        validate_assignment = True
        extra = "allow"

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            load_dotenv()
            return env_settings, file_secret_settings, init_settings

    @root_validator()
    def _check_and_set_values(cls, values):
        for k, v in values.items():
            if v is not None:
                old_value = os.getenv(k.upper())
                if old_value is None or old_value != str(v):
                    os.environ[k.upper()] = str(v)
                    logger.debug(f"Set environment variable {k.upper()}={v}")
        return values



import os
from pathlib import Path
from typing import Any, Union

from pydantic import BaseModel, validator

from hyfi.__global__ import __about__
from hyfi.dotenv import DotEnvConfig
from hyfi.hydra import _compose
from hyfi.joblib import JobLibConfig
from hyfi.path import PathConfig
from hyfi.utils.logging import getLogger
from hyfi.utils.notebook import is_notebook

logger = getLogger(__name__)


class ProjectConfig(BaseModel):
    """Project Config"""

    config_name: str = "__init__"
    project_name: str = "hyfi-project"
    task_name: str = ""
    project_description: str = ""
    project_root: str = ""
    project_workspace_name: str = "workspace"
    global_hyfi_root: str = ""
    global_workspace_name: str = "workspace"
    num_workers: int = 1
    use_huggingface_hub: bool = False
    use_wandb: bool = False
    verbose: Union[bool, int] = False
    # Config Classes
    dotenv: DotEnvConfig = None  # type: ignore
    joblib: JobLibConfig = None  # type: ignore
    path: PathConfig = None  # type: ignore

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True

    @validator("project_name", allow_reuse=True)
    def _validate_project_name(cls, v):
        if v is None:
            raise ValueError("Project name must be specified.")
        return v

    @validator("verbose", allow_reuse=True)
    def _validate_verbose(cls, v):
        if isinstance(v, str):
            if v.lower() in {"true", "1"}:
                v = True
            elif v.lower() in {"false", "0"}:
                v = False
            else:
                raise ValueError("verbose must be a boolean or a string of 0 or 1")
        return v

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        data = _compose(
            f"project={config_name}",
            config_data=data,
            config_module=__about__.config_module,
        )  # type: ignore
        super().__init__(config_name=config_name, **data)

    def init_project(self):
        self.dotenv = DotEnvConfig()
        if self.path is None:
            self.path = PathConfig()
        if self.joblib is None:
            self.joblib = JobLibConfig()

        self.dotenv.HYFI_PROJECT_NAME = self.project_name
        self.dotenv.HYFI_TASK_NAME = self.task_name
        self.dotenv.HYFI_PROJECT_DESC = self.project_description
        self.dotenv.HYFI_PROJECT_ROOT = self.project_root
        self.dotenv.HYFI_PROJECT_WORKSPACE_NAME = self.project_workspace_name
        self.dotenv.HYFI_GLOBAL_ROOT = self.global_hyfi_root
        self.dotenv.HYFI_GLOBAL_WORKSPACE_NAME = self.global_workspace_name
        self.dotenv.HYFI_NUM_WORKERS = self.num_workers
        self.dotenv.HYFI_VERBOSE = self.verbose
        self.dotenv.CACHED_PATH_CACHE_ROOT = str(self.path.cache_dir / "cached_path")
        self.init_wandb()
        if self.use_huggingface_hub:
            self.init_huggingface_hub()

    def init_wandb(self):
        if self.path is None:
            raise ValueError("Path object not initialized")
        if self.dotenv is None:
            raise ValueError("DotEnv object not initialized")

        self.dotenv.WANDB_DIR = str(self.path.log_dir)
        project_name = self.project_name.replace("/", "-").replace("\\", "-")
        self.dotenv.WANDB_PROJECT = project_name
        task_name = self.task_name.replace("/", "-").replace("\\", "-")
        notebook_name = self.path.log_dir / f"{task_name}-nb"
        notebook_name.mkdir(parents=True, exist_ok=True)
        self.dotenv.WANDB_NOTEBOOK_NAME = str(notebook_name)
        self.dotenv.WANDB_SILENT = str(not self.verbose)
        if self.use_wandb:
            try:
                import wandb  # type: ignore

                wandb.init(project=self.project_name)
            except ImportError:
                logger.warning(
                    "wandb is not installed, please install it to use wandb."
                )

    def init_huggingface_hub(self):
        """Initialize huggingface_hub"""
        try:
            from huggingface_hub import notebook_login  # type: ignore
            from huggingface_hub.hf_api import HfFolder  # type: ignore
        except ImportError:
            logger.warning(
                "huggingface_hub is not installed, please install it to use huggingface_hub."
            )
            return

        self.dotenv = DotEnvConfig()
        if (
            self.dotenv.HUGGING_FACE_HUB_TOKEN is None
            and self.dotenv.HF_USER_ACCESS_TOKEN is not None
        ):
            self.dotenv.HUGGING_FACE_HUB_TOKEN = self.dotenv.HF_USER_ACCESS_TOKEN

        local_token = HfFolder.get_token()
        if local_token is None:
            if is_notebook():
                notebook_login()
            else:
                logger.info(
                    "huggingface_hub.notebook_login() is only available in notebook,"
                    "set HUGGING_FACE_HUB_TOKEN manually"
                )

    @property
    def environ(self):
        return os.environ

    @property
    def project_workspace_dir(self):
        if self.path is None:
            raise ValueError("Path object not initialized")
        _p = Path(self.path.project_workspace_root)
        _p.mkdir(parents=True, exist_ok=True)
        return _p.absolute()

    @property
    def project_dir(self):
        if self.path is None:
            raise ValueError("Path object not initialized")
        _p = Path(self.path.project_root)
        _p.mkdir(parents=True, exist_ok=True)
        return _p.absolute()



"""
    This module contains the primary class for the hyfi config package, HyFI,
    as well as various utility functions and imports.
"""
from pathlib import Path
from typing import IO, Any, Dict, List, Tuple, Union

from omegaconf import DictConfig, ListConfig, SCMode

from hyfi.__global__ import __home_path__, __hyfi_path__
from hyfi.__global__.config import __global_config__
from hyfi.cached_path import cached_path
from hyfi.dotenv import DotEnvConfig
from hyfi.hydra import _compose, _select, _to_config, _to_dict
from hyfi.hydra.main import (
    DictKeyType,
    SpecialKeys,
    _ensure_kwargs,
    _ensure_list,
    _function,
    _getsource,
    _instantiate,
    _is_config,
    _is_instantiatable,
    _is_list,
    _load,
    _load_json,
    _merge,
    _methods,
    _partial,
    _print,
    _run,
    _save,
    _save_json,
    _to_container,
    _to_yaml,
    _update,
    _viewsource,
)
from hyfi.joblib.pipe import _apply, _pipe
from hyfi.project import ProjectConfig
from hyfi.utils.env import expand_posix_vars, get_osenv, load_dotenv, set_osenv
from hyfi.utils.file import exists, is_dir, is_file, join_path, mkdir
from hyfi.utils.func import (
    dict_product,
    dict_to_dataframe,
    records_to_dataframe,
    to_dateparm,
    to_datetime,
    to_numeric,
)
from hyfi.utils.google import mount_google_drive
from hyfi.utils.gpu import nvidia_smi, set_cuda
from hyfi.utils.logging import getLogger, setLogger
from hyfi.utils.notebook import (
    clear_output,
    cprint,
    create_button,
    create_dropdown,
    create_floatslider,
    create_image,
    create_radiobutton,
    create_textarea,
    display,
    display_image,
    get_display,
    hide_code_in_slideshow,
    is_colab,
    is_notebook,
)

logger = getLogger(__name__)


def _about(cfg):
    pkg_name = cfg.about.__package_name__
    name = cfg.about.name
    print()
    for k, v in cfg.about.dict().items():
        if k.startswith("_"):
            continue
        print(f"{k:11} : {v}")
    if pkg_name:
        print(f"\nExecute `{pkg_name} --help` to see what you can do with {name}")


class HyFI:
    """Primary class for the hyfi config package"""

    config = __global_config__
    SpeicialKeys = SpecialKeys
    __version__ = __global_config__.__version__
    __hyfi_path__ = __hyfi_path__()
    __home_path__ = __home_path__()

    def __init__(self) -> None:
        raise NotImplementedError("Use one of the static construction functions")

    @staticmethod
    def about() -> None:
        """Print the about information"""
        cfg = __global_config__
        _about(cfg)

    @staticmethod
    def initialize(config: Union[DictConfig, Dict] = None):  # type: ignore
        """Initialize the global config"""
        __global_config__.initialize(config)

    @staticmethod
    def terminate():
        """Terminate the global config"""
        __global_config__.terminate()

    @staticmethod
    def envs() -> DotEnvConfig:
        """Return the current environments"""
        return DotEnvConfig()  # type: ignore

    @staticmethod
    def expand_posix_vars(posix_expr: str, context: dict = None) -> str:  # type: ignore
        """
        Expand POSIX variables in a string.

        Args:
            posix_expr (str): The string containing POSIX variables to be expanded.
            context (dict, optional): A dictionary containing additional variables to be used in the expansion.
                Defaults to None.

        Returns:
            str: The expanded string.

        Examples:
            >>> expand_posix_vars("$HOME")
            '/home/user'
            >>> expand_posix_vars("$HOME/$USER", {"USER": "testuser"})
            '/home/user/testuser'

        """
        return expand_posix_vars(posix_expr, context=context)

    @staticmethod
    def environ(key: str = "", default: Union[str, None] = None) -> Any:
        """Get the value of an environment variable or return the default value"""
        return get_osenv(key, default=default)

    @staticmethod
    def compose(
        config_group: Union[str, None] = None,
        overrides: Union[List[str], None] = None,
        config_data: Union[Dict[str, Any], DictConfig, None] = None,
        *,
        return_as_dict: bool = True,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
        config_name: Union[str, None] = None,
        config_module: Union[str, None] = None,
        global_package: bool = False,
        verbose: bool = False,
    ) -> Union[DictConfig, Dict]:
        """
        Compose a configuration by applying overrides

        Args:
            config_group: Name of the config group to compose (`config_group=name`)
            overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
            config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group.key=value`)
            return_as_dict: Return the result as a dict
            throw_on_resolution_failure: If True throw an exception if resolution fails
            throw_on_missing: If True throw an exception if config_group doesn't exist
            config_name: Name of the root config to be used (e.g. `hconf`)
            config_module: Module of the config to be used (e.g. `hyfi.conf`)
            global_package: If True, the config assumed to be a global package
            verbose: If True print configuration to stdout

        Returns:
            A config object or a dictionary with the composed config
        """
        return _compose(
            config_group=config_group,
            overrides=overrides,
            config_data=config_data,
            return_as_dict=return_as_dict,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
            config_name=config_name,
            config_module=config_module,
            global_package=global_package,
            verbose=verbose,
        )

    @staticmethod
    def select(
        cfg: Any,
        key: str,
        *,
        default: Any = None,
        throw_on_resolution_failure: bool = True,
        throw_on_missing: bool = False,
    ):
        return _select(
            cfg,
            key,
            default=default,
            throw_on_resolution_failure=throw_on_resolution_failure,
            throw_on_missing=throw_on_missing,
        )

    @staticmethod
    def to_dict(
        cfg: Any,
    ):
        return _to_dict(cfg)

    @staticmethod
    def to_config(
        cfg: Any,
    ):
        return _to_config(cfg)

    @staticmethod
    def to_yaml(cfg: Any, *, resolve: bool = False, sort_keys: bool = False) -> str:
        if resolve:
            cfg = _to_dict(cfg)
        return _to_yaml(cfg, resolve=resolve, sort_keys=sort_keys)

    @staticmethod
    def to_container(
        cfg: Any,
        *,
        resolve: bool = False,
        throw_on_missing: bool = False,
        enum_to_str: bool = False,
        structured_config_mode: SCMode = SCMode.DICT,
    ):
        return _to_container(
            cfg=cfg,
            resolve=resolve,
            throw_on_missing=throw_on_missing,
            enum_to_str=enum_to_str,
            structured_config_mode=structured_config_mode,
        )

    @staticmethod
    def partial(
        config: Any = None,
        config_group: Union[str, None] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return _partial(config=config, config_group=config_group, *args, **kwargs)

    @staticmethod
    def instantiate(config: Any, *args: Any, **kwargs: Any) -> Any:
        return _instantiate(config, *args, **kwargs)

    @staticmethod
    def is_config(
        cfg: Any,
    ):
        return _is_config(cfg)

    @staticmethod
    def is_list(
        cfg: Any,
    ):
        return _is_list(cfg)

    @staticmethod
    def is_instantiatable(cfg: Any):
        return _is_instantiatable(cfg)

    @staticmethod
    def load(file_: Union[str, Path, IO[Any]]) -> Union[DictConfig, ListConfig]:
        return _load(file_)

    @staticmethod
    def update(_dict, _overrides):
        return _update(_dict, _overrides)

    @staticmethod
    def merge(
        *configs: Union[
            DictConfig,
            ListConfig,
            Dict[DictKeyType, Any],
            List[Any],
            Tuple[Any, ...],
            Any,
        ],
    ) -> Union[ListConfig, DictConfig]:
        """
        Merge a list of previously created configs into a single one
        :param configs: Input configs
        :return: the merged config object.
        """
        return _merge(*configs)

    @staticmethod
    def save(config: Any, f: Union[str, Path, IO[Any]], resolve: bool = False) -> None:
        _save(config, f, resolve)

    @staticmethod
    def save_json(
        json_dict: dict,
        f: Union[str, Path, IO[Any]],
        indent=4,
        ensure_ascii=False,
        default=None,
        **kwargs,
    ):
        _save_json(json_dict, f, indent, ensure_ascii, default, **kwargs)

    @staticmethod
    def load_json(f: Union[str, Path, IO[Any]], **kwargs) -> dict:
        return _load_json(f, **kwargs)

    @staticmethod
    def pprint(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def print(cfg: Any, resolve: bool = True, **kwargs):
        _print(cfg, resolve=resolve, **kwargs)

    @staticmethod
    def methods(cfg: Any, obj: object, return_function=False):
        return _methods(cfg, obj, return_function)

    @staticmethod
    def function(cfg: Any, _name_, return_function=False, **parms):
        return _function(cfg, _name_, return_function, **parms)

    @staticmethod
    def run(config: Any, **kwargs: Any) -> Any:
        _run(config, **kwargs)

    @staticmethod
    def load_dotenv(
        override: bool = False,
        dotenv_dir: str = "",
        dotenv_filename: str = ".env",
        verbose: bool = False,
    ) -> None:
        load_dotenv(
            override=override,
            dotenv_filename=dotenv_filename,
            dotenv_dir=dotenv_dir,
            verbose=verbose,
        )

    @staticmethod
    def cached_path(
        url_or_filename,
        extract_archive: bool = False,
        force_extract: bool = False,
        return_parent_dir: bool = False,
        cache_dir=None,
        verbose: bool = False,
    ):
        """
        Attempts to cache a file or URL and return the path to the cached file.
        If required libraries 'cached_path' and 'gdown' are not installed, raises an ImportError.

        Args:
            url_or_filename (str): The URL or filename to be cached.
            extract_archive (bool, optional): Whether to extract the file if it's an archive. Defaults to False.
            force_extract (bool, optional): Whether to force extraction even if the destination already exists. Defaults to False.
            return_parent_dir (bool, optional): If True, returns the parent directory of the cached file. Defaults to False.
            cache_dir (str, optional): Directory to store cached files. Defaults to None.
            verbose (bool, optional): Whether to print informative messages during the process. Defaults to False.

        Raises:
            ImportError: If the required libraries 'cached_path' and 'gdown' are not imported.

        Returns:
            str: Path to the cached file or its parent directory, depending on the 'return_parent_dir' parameter.
        """
        return cached_path(
            url_or_filename,
            extract_archive=extract_archive,
            force_extract=force_extract,
            return_parent_dir=return_parent_dir,
            cache_dir=cache_dir,
            verbose=verbose,
        )

    @staticmethod
    def pipe(data=None, cfg=None):
        return _pipe(data, cfg)

    @staticmethod
    def ensure_list(value):
        return _ensure_list(value)

    @staticmethod
    def to_dateparm(_date, _format="%Y-%m-%d"):
        return to_dateparm(_date, _format)

    @staticmethod
    def exists(a, *p):
        return exists(a, *p)

    @staticmethod
    def is_file(a, *p):
        return is_file(a, *p)

    @staticmethod
    def is_dir(a, *p):
        return is_dir(a, *p)

    @staticmethod
    def mkdir(_path: str):
        return mkdir(_path)

    @staticmethod
    def join_path(a, *p):
        return join_path(a, *p)

    @staticmethod
    def apply(
        func,
        series,
        description=None,
        use_batcher=True,
        minibatch_size=None,
        num_workers=None,
        verbose=False,
        **kwargs,
    ):
        return _apply(
            func,
            series,
            description=description,
            use_batcher=use_batcher,
            minibatch_size=minibatch_size,
            num_workers=num_workers,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def ensure_kwargs(_kwargs, _fn):
        return _ensure_kwargs(_kwargs, _fn)

    @staticmethod
    def save_data(
        data,
        filename=None,
        base_dir=None,
        columns=None,
        index=False,
        filetype="parquet",
        suffix=None,
        verbose=False,
        **kwargs,
    ):
        from hyfi.utils.file import save_data

        if filename is None:
            raise ValueError("filename must be specified")
        save_data(
            data,
            filename,
            base_dir=base_dir,
            columns=columns,
            index=index,
            filetype=filetype,
            suffix=suffix,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def load_data(filename=None, base_dir=None, filetype=None, verbose=False, **kwargs):
        from hyfi.utils.file import load_data

        if filename is not None:
            filename = str(filename)
        if SpecialKeys.TARGET in kwargs:
            return _instantiate(
                kwargs,
                filename=filename,
                base_dir=base_dir,
                verbose=verbose,
                filetype=filetype,
            )
        if filename is None:
            raise ValueError("filename must be specified")
        return load_data(
            filename,
            base_dir=base_dir,
            verbose=verbose,
            filetype=filetype,
            **kwargs,
        )

    @staticmethod
    def get_filepaths(
        filename_patterns=None, base_dir=None, recursive=True, verbose=True, **kwargs
    ):
        from hyfi.utils.file import get_filepaths

        if filename_patterns is None:
            raise ValueError("filename must be specified")
        return get_filepaths(
            filename_patterns,
            base_dir=base_dir,
            recursive=recursive,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def concat_data(
        data,
        columns=None,
        add_key_as_name=False,
        name_column="_name_",
        ignore_index=True,
        verbose=False,
        **kwargs,
    ):
        from hyfi.utils.file import concat_data

        return concat_data(
            data,
            columns=columns,
            add_key_as_name=add_key_as_name,
            name_column=name_column,
            ignore_index=ignore_index,
            verbose=verbose,
            **kwargs,
        )

    @staticmethod
    def is_dataframe(data):
        from hyfi.utils.file import is_dataframe

        return is_dataframe(data)

    @staticmethod
    def is_colab():
        return is_colab()

    @staticmethod
    def is_notebook():
        return is_notebook()

    @staticmethod
    def osenv(key, default=None):
        return get_osenv(key, default)

    @staticmethod
    def env_set(key, value):
        return set_osenv(key, value)

    @staticmethod
    def nvidia_smi():
        return nvidia_smi()

    @staticmethod
    def ensure_import_module(name, libpath, liburi, specname=None, syspath=None):
        from hyfi.utils.lib import ensure_import_module

        return ensure_import_module(name, libpath, liburi, specname, syspath)

    @staticmethod
    def collage(
        images_or_uris,
        collage_filepath=None,
        ncols=3,
        max_images=12,
        collage_width=1200,
        padding: int = 10,
        bg_color: str = "black",
        crop_to_min_size=False,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=12,
        fontcolor="#000",
        **kwargs,
    ):
        from hyfi.graphic.collage import collage as _collage

        return _collage(
            images_or_uris,
            collage_filepath=collage_filepath,
            ncols=ncols,
            max_images=max_images,
            collage_width=collage_width,
            padding=padding,
            bg_color=bg_color,
            crop_to_min_size=crop_to_min_size,
            show_filename=show_filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
            **kwargs,
        )

    @staticmethod
    def make_gif(
        image_filepaths=None,
        filename_patterns=None,
        base_dir=None,
        output_filepath=None,
        duration=100,
        loop=0,
        width=None,
        optimize=True,
        quality=50,
        show=False,
        force=False,
        **kwargs,
    ):
        from hyfi.graphic.motion import make_gif as _make_gif

        return _make_gif(
            image_filepaths=image_filepaths,
            filename_patterns=filename_patterns,
            base_dir=base_dir,
            output_filepath=output_filepath,
            duration=duration,
            loop=loop,
            width=width,
            optimize=optimize,
            quality=quality,
            show=show,
            force=force,
            **kwargs,
        )

    @staticmethod
    def to_datetime(data, _format=None, _columns=None, **kwargs):
        return to_datetime(data, _format, _columns, **kwargs)

    @staticmethod
    def to_numeric(data, _columns=None, errors="coerce", downcast=None, **kwargs):
        return to_numeric(data, _columns, errors, downcast, **kwargs)

    @staticmethod
    def getLogger(
        name=None,
        log_level=None,
        fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    ):
        return getLogger(name, log_level, fmt)

    @staticmethod
    def setLogger(level=None, force=True, **kwargs):
        return setLogger(level, force, **kwargs)

    @staticmethod
    def set_cuda(device=0):
        return set_cuda(device)

    @staticmethod
    def mount_google_drive(
        project_root: str = "",
        project_name: str = "",
        mountpoint: str = "/content/drive",
        force_remount: bool = False,
        timeout_ms: int = 120000,
    ):
        return mount_google_drive(
            project_root, project_name, mountpoint, force_remount, timeout_ms
        )

    @staticmethod
    def getsource(obj):
        return _getsource(obj)

    @staticmethod
    def viewsource(obj):
        return _viewsource(obj)

    @staticmethod
    def clear_output(wait=False):
        return clear_output(wait)

    @staticmethod
    def display(
        *objs,
        include=None,
        exclude=None,
        metadata=None,
        transient=None,
        display_id=None,
        **kwargs,
    ):
        return display(
            *objs,
            include=include,
            exclude=exclude,
            metadata=metadata,
            transient=transient,
            display_id=display_id,
            **kwargs,
        )

    @staticmethod
    def display_image(
        data=None,
        url=None,
        filename=None,
        format=None,
        embed=None,
        width=None,
        height=None,
        retina=False,
        unconfined=False,
        metadata=None,
        **kwargs,
    ):
        return display_image(
            data=data,
            url=url,
            filename=filename,
            format=format,
            embed=embed,
            width=width,
            height=height,
            retina=retina,
            unconfined=unconfined,
            metadata=metadata,
            **kwargs,
        )

    @staticmethod
    def pip(
        name,
        upgrade=False,
        prelease=False,
        editable=False,
        quiet=False,
        find_links=None,
        requirement=None,
        force_reinstall=False,
        verbose=False,
        **kwargs,
    ):
        from hyfi.utils.lib import pip as _pip

        return _pip(
            name,
            upgrade,
            prelease,
            editable,
            quiet,
            find_links,
            requirement,
            force_reinstall,
            verbose,
            **kwargs,
        )

    @staticmethod
    def upgrade(prelease=False, quiet=False, force_reinstall=False, **kwargs):
        from hyfi.utils.lib import pip

        return pip(
            name="hyfi",
            upgrade=True,
            prelease=prelease,
            quiet=quiet,
            force_reinstall=force_reinstall,
            **kwargs,
        )

    @staticmethod
    def dict_product(dicts):
        return dict_product(dicts)

    @staticmethod
    def get_display():
        return get_display()

    @staticmethod
    def hide_code_in_slideshow():
        return hide_code_in_slideshow()

    @staticmethod
    def cprint(str_color_tuples, **kwargs):
        return cprint(str_color_tuples)

    @staticmethod
    def dict_to_dataframe(data, orient="columns", dtype=None, columns=None):
        return dict_to_dataframe(data, orient, dtype, columns)

    @staticmethod
    def records_to_dataframe(
        data, index=None, exclude=None, columns=None, coerce_float=False, nrows=None
    ):
        return records_to_dataframe(data, index, exclude, columns, coerce_float, nrows)

    @staticmethod
    def create_dropdown(
        options, value, description, disabled=False, style=None, layout=None, **kwargs
    ):
        if style is None:
            style = {"description_width": "initial"}
        return create_dropdown(
            options,
            value,
            description,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_textarea(
        value,
        description,
        placeholder="",
        disabled=False,
        style=None,
        layout=None,
        **kwargs,
    ):
        if style is None:
            style = {"description_width": "initial"}
        return create_textarea(
            value,
            description,
            placeholder,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_button(
        description, button_style="", icon="check", layout=None, **kwargs
    ):
        return create_button(description, button_style, icon, layout, **kwargs)

    @staticmethod
    def create_radiobutton(
        options,
        description,
        value=None,
        disabled=False,
        style=None,
        layout=None,
        **kwargs,
    ):
        if style is None:
            style = {"description_width": "initial"}
        return create_radiobutton(
            options,
            description,
            value,
            disabled,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def create_image(
        filename=None,
        format=None,
        width=None,
        height=None,
        **kwargs,
    ):
        return create_image(filename, format, width, height, **kwargs)

    @staticmethod
    def create_floatslider(
        min=0.0,
        max=1.0,
        step=0.1,
        value=None,
        description="",
        disabled=False,
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format=".1f",
        style=None,
        layout=None,
        **kwargs,
    ):
        if style is None:
            style = {"description_width": "initial"}
        return create_floatslider(
            min,
            max,
            step,
            value,
            description,
            disabled,
            continuous_update,
            orientation,
            readout,
            readout_format,
            style,
            layout,
            **kwargs,
        )

    @staticmethod
    def get_image_font(fontname=None, fontsize=12):
        from hyfi.graphic.collage import get_image_font

        return get_image_font(fontname, fontsize)

    @staticmethod
    def read(uri, mode="rb", encoding=None, head=None, **kwargs):
        from hyfi.utils.file import read as _read

        return _read(uri, mode, encoding, head, **kwargs)

    @staticmethod
    def load_image(
        image_or_uri,
        max_width: int = None,
        max_height: int = None,
        max_pixels: int = None,
        scale: float = 1.0,
        resize_to_multiple_of: int = None,
        crop_box=None,
        mode="RGB",
        **kwargs,
    ):
        from hyfi.graphic.utils import load_image as _load_image

        return _load_image(
            image_or_uri,
            max_width,
            max_height,
            max_pixels,
            scale,
            resize_to_multiple_of,
            crop_box,
            mode,
            **kwargs,
        )

    @staticmethod
    def init_workspace(
        project_name: str = "",
        task_name: str = "",
        project_description: str = "",
        project_root: str = "",
        project_workspace_name: str = "",
        global_hyfi_root: str = "",
        global_workspace_name: str = "",
        num_workers: int = -1,
        log_level: str = "",
        autotime: bool = True,
        retina: bool = True,
        verbose: Union[bool, int] = False,
        **kwargs,
    ) -> ProjectConfig:
        """
        Initialize and start hyfi.

        Args:
                project_name: Name of the project to use.
                task_name: Name of the task to use.
                project_description: Description of the project that will be used.
                project_root: Root directory of the project.
                project_workspace_name: Name of the project's workspace directory.
                global_hyfi_root: Root directory of the global hyfi.
                global_workspace_name: Name of the global hierachical workspace directory.
                num_workers: Number of workers to run.
                log_level: Log level for the log.
                autotime: Whether to automatically set time and / or keep track of run times.
                retina: Whether to use retina or not.
                verbose: Enables or disables logging
        """
        __global_config__.init_workspace(
            project_name=project_name,
            task_name=task_name,
            project_description=project_description,
            project_root=project_root,
            project_workspace_name=project_workspace_name,
            global_hyfi_root=global_hyfi_root,
            global_workspace_name=global_workspace_name,
            num_workers=num_workers,
            log_level=log_level,
            autotime=autotime,
            retina=retina,
            verbose=verbose,
            **kwargs,
        )
        if __global_config__.project:
            return __global_config__.project
        else:
            raise ValueError("Project not initialized.")

    @staticmethod
    def scale_image(
        image,
        max_width: int = None,
        max_height: int = None,
        max_pixels: int = None,
        scale: float = 1.0,
        resize_to_multiple_of: int = 8,
        resample: int = 1,
    ):
        """
        Scale an image to a maximum width, height, or number of pixels.

        resample:   Image.NEAREST (0), Image.LANCZOS (1), Image.BILINEAR (2),
                    Image.BICUBIC (3), Image.BOX (4) or Image.HAMMING (5)
        """
        from hyfi.graphic.utils import scale_image as _scale_image

        return _scale_image(
            image,
            max_width,
            max_height,
            max_pixels,
            scale,
            resize_to_multiple_of,
            resample,
        )

    @staticmethod
    def copy(src, dst, *, follow_symlinks=True):
        """
        Copy a file or directory. This is a wrapper around shutil.copy.
        If you need to copy an entire directory (including all of its contents), or if you need to overwrite existing files in the destination directory, shutil.copy() would be a better choice.

        Args:
                src: Path to the file or directory to be copied.
                dst: Path to the destination directory. If the destination directory does not exist it will be created.
                follow_symlinks: Whether or not symlinks should be followed
        """
        import shutil

        src = str(src)
        dst = str(dst)
        mkdir(dst)
        shutil.copy(src, dst, follow_symlinks=follow_symlinks)
        logger.info(f"copied {src} to {dst}")

    @staticmethod
    def copyfile(src, dst, *, follow_symlinks=True):
        """
        Copy a file or directory. This is a wrapper around shutil.copyfile.
        If you want to copy a single file from one location to another, shutil.copyfile() is the appropriate function to use.

        Args:
                src: Path to the file or directory to copy.
                dst: Path to the destination file or directory. If the destination file already exists it will be overwritten.
                follow_symlinks: Whether to follow symbolic links or not
        """
        import shutil

        src = str(src)
        dst = str(dst)
        shutil.copyfile(src, dst, follow_symlinks=follow_symlinks)
        logger.info(f"copied {src} to {dst}")

    @staticmethod
    def gpu_usage(all=False, attrList=None, useOldCode=False):
        """
        Show GPU utilization in human readable format. This is a wrapper around the GPUtil library.

        Args:
                all: If True show all available GPUs ( default : False )
                attrList: List of attributes to show ( default : None )
                useOldCode: If True use old code instead of new code ( default : False )

        Returns:
                A string with the
        """
        try:
            from GPUtil import showUtilization  # type: ignore
        except ImportError:
            logger.error("GPUtil is not installed. To install, run: pip install GPUtil")
            return

        return showUtilization(all, attrList, useOldCode)



import contextlib
from pathlib import Path
from typing import Union

from omegaconf import DictConfig
from pydantic import BaseModel

from hyfi.hydra import _compose, _to_config
from hyfi.hydra.main import _merge, _methods, _print
from hyfi.module import ModuleConfig
from hyfi.path.batch import BatchPathConfig
from hyfi.project import ProjectConfig
from hyfi.utils.lib import ensure_import_module
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class BaseTask(BaseModel):
    config_name: str = "__init__"
    config_group: str = "task"
    task_name: str
    path: BatchPathConfig = None  # type: ignore
    project: ProjectConfig = None  # type: ignore
    module: ModuleConfig = None  # type: ignore
    autoload: bool = False
    version: str = "0.0.0"
    _config_: DictConfig = None  # type: ignore
    _initial_config_: DictConfig = None  # type: ignore

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
        validate_assignment = False
        exclude = {
            "_config_",
            "_initial_config_",
            "__data__",
            "path",
            "module",
            "project",
        }
        include = {}
        underscore_attrs_are_private = True
        property_set_methods = {
            "name": "set_name",
            "root_dir": "set_root_dir",
        }

    def __init__(
        self,
        config_name: str = "",
        config_group: str = "",
        root_dir: str = "",
        **args,
    ):
        if config_group:
            args = _merge(_compose(config_group), args)
        else:
            args = _to_config(args)
        super().__init__(config_name=config_name, config_group=config_group, **args)

        object.__setattr__(self, "_config_", args)
        object.__setattr__(self, "_initial_config_", args.copy())
        self.initialize_configs(root_dir=root_dir)

    def __setattr__(self, key, val):
        super().__setattr__(key, val)
        method = self.__config__.property_set_methods.get(key)
        if method is not None:
            getattr(self, method)(val)

    def set_root_dir(self, root_dir: Union[str, Path]):
        path = self.config.path
        if path is None:
            path = _compose("path=_batch_")
            logger.info(
                f"There is no path in the config, using default path: {path.root}"
            )
            self._config_.path = path
        if root_dir is not None:
            path.root = str(root_dir)
        self.path = BatchPathConfig(**path)

    def set_name(self, val):
        self._config_.name = val
        if self.task_name is None or self.task_name != val:
            self.task_name = val

    def initialize_configs(self, root_dir=None, **kwargs):
        self.root_dir = root_dir

    @property
    def config(self):
        return self._config_

    @property
    def root_dir(self) -> Path:
        return Path(self.path.root)

    @property
    def output_dir(self):
        return self.path.output_dir

    @property
    def project_name(self):
        return self.project.project_name

    @property
    def project_dir(self):
        return Path(self.project.project_dir)

    @property
    def workspace_dir(self):
        return Path(self.project.project_workspace_dir)

    @property
    def model_dir(self):
        return self.path.model_dir

    @property
    def log_dir(self):
        return self.project.path.log_dir

    @property
    def cache_dir(self):
        cache_dir = Path(self.project.path.global_cache)
        if cache_dir is None:
            cache_dir = self.output_dir / ".cache"
            cache_dir.mkdir(parents=True, exist_ok=True)
        return Path(cache_dir)

    @property
    def library_dir(self):
        return self.path.library_dir

    @property
    def verbose(self):
        return self.project.verbose

    def autorun(self):
        return _methods(self.auto, self)

    def show_config(self):
        _print(self.dict())

    def load_modules(self):
        """Load the modules"""
        if self.module.get("modules") is None:
            logger.info("No modules to load")
            return
        library_dir = self.library_dir
        for module in self.module.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = library_dir / libname
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = library_dir / syspath
            ensure_import_module(name, libpath, liburi, specname, syspath)

    def reset(self, objects=None):
        """Reset the memory cache"""
        if isinstance(objects, list):
            for obj in objects:
                del obj
        with contextlib.suppress(ImportError):
            from hyfi.utils.gpu import GPUMon

            GPUMon.release_gpu_memory()



from tqdm.auto import tqdm

from hyfi.hydra.main import SpecialKeys, _partial
from hyfi.joblib.batch import batcher
from hyfi.joblib.batch.apply import decorator_apply
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def _pipe(data, pipe):
    _func_ = pipe.get(SpecialKeys.FUNC)
    _fn = _partial(_func_)
    logger.info("Applying pipe: %s", _fn)
    if isinstance(data, dict):
        if "concat_dataframes" in str(_fn):
            return _fn(data, pipe)
        dfs = {}
        for df_no, df_name in enumerate(data):
            df_each = data[df_name]

            logger.info(
                "Applying pipe to dataframe [%s], %d/%d", df_name, df_no + 1, len(data)
            )

            pipe[SpecialKeys.SUFFIX.value] = df_name
            dfs[df_name] = _fn(df_each, pipe)
        return dfs
    return _fn(data, pipe)


def _apply(
    func,
    series,
    description=None,
    use_batcher=True,
    minibatch_size=None,
    num_workers=None,
    **kwargs,
):
    batcher_instance = batcher.batcher_instance
    if use_batcher and batcher_instance is not None:
        batcher_minibatch_size = batcher_instance.minibatch_size
        if minibatch_size is None:
            minibatch_size = batcher_minibatch_size
        if num_workers is not None:
            batcher_instance.procs = int(num_workers)
        if batcher_instance.procs > 1:
            batcher_instance.minibatch_size = min(
                int(len(series) / batcher_instance.procs) + 1, minibatch_size
            )
            logger.info(
                f"Using batcher with minibatch size: {batcher_instance.minibatch_size}"
            )
            results = decorator_apply(func, batcher_instance, description=description)(
                series
            )
            if batcher_instance is not None:
                batcher_instance.minibatch_size = batcher_minibatch_size
            return results

    if batcher_instance is None:
        logger.info("Warning: batcher not initialized")
    tqdm.pandas(desc=description)
    return series.progress_apply(func)



"""HyFI: Hydra Fast Interface (Hydra and Pydantic based interface framework)"""
from hyfi.__cli__ import hydra_main
from hyfi.__global__ import __about__ as about
from hyfi.__global__ import __hydra_version_base__
from hyfi.__global__.config import __global_config__ as global_config
from hyfi.main import HyFI
from hyfi.main import HyFI as H
from hyfi.main import HyFI as HI
from hyfi.utils.logging import getLogger



"""
Hyfi configuration file.
"""
from typing import Any, Dict, Optional, Union

from omegaconf import DictConfig
from pydantic import BaseModel, root_validator, validator

from hyfi.__global__ import __about__, __hydra_config__
from hyfi.about import AboutConfig
from hyfi.dotenv import DotEnvConfig
from hyfi.hydra import _compose
from hyfi.project import ProjectConfig
from hyfi.utils.env import _check_and_set_value, expand_posix_vars
from hyfi.utils.logging import getLogger, setLogger
from hyfi.utils.notebook import load_extentions, set_matplotlib_formats

logger = getLogger(__name__)


def __version__():
    """
    Returns the version of Hyfi. It is used to determine the version of Hyfi.


    Returns:
        string containing the version of
    """
    from hyfi._version import __version__

    return __version__


class HyfiConfig(BaseModel):
    """HyFI root config class.  This class is used to store the configuration"""

    hyfi_config_path: str = __about__.config_path
    hyfi_config_module: str = __about__.config_module
    hyfi_user_config_path: str = ""

    debug_mode: bool = False
    print_config: bool = False
    print_resolved_config: bool = False
    verbose: bool = False
    logging_level: str = "WARNING"

    hydra: Optional[DictConfig] = None

    about: AboutConfig = AboutConfig()
    project: Optional[ProjectConfig] = None
    copier: Optional[DictConfig] = None

    __version__: str = __version__()
    __initilized__: bool = False

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        validate_assignment = True
        extra = "allow"

    @root_validator()
    def _check_and_set_values(cls, values):
        """
        Validate and set values for the config file.

        Args:
                cls: Class to use for config lookup
                values: Dictionary of values to check and set

        Returns:
                Same dictionary with hyfi_config
        """
        key = "hyfi_config_path"
        val = _check_and_set_value(key, values.get(key))
        values[key] = val
        # Set the hyfi_config_module value in the configuration file.
        if val is not None:
            key = "hyfi_config_module"
            values[key] = _check_and_set_value(key, val.replace("pkg://", ""))
        return values

    @validator("hyfi_user_config_path")
    def _validate_hyfi_user_config_path(cls, v):
        """
        Validate and set hyfi_user_config_path.

        Args:
                cls: Class to use for validation.
                v: Value to set if valid.

        Returns:
                True if valid False otherwise
        """
        return _check_and_set_value("hyfi_user_config_path", v)

    @validator("logging_level")
    def _validate_logging_level(cls, v, values):
        """
        Validate and set the logging level

        Args:
                cls: The class to operate on
                v: The value to set the logging level to
                values: The values from the config file

        Returns:
                The value that was set
        """
        verbose = values.get("verbose", False)
        # Set verbose to INFO.
        if verbose and v == "WARNING":
            v = "INFO"
        logger.setLevel(v)
        return v

    def __init__(self, **data: Any):
        """
        Initialize the object with data

        Args:
                data: Data to initialize the
        """
        super().__init__(**data)
        # self.about = __about__

    def init_workspace(
        self,
        project_name: str = "",
        task_name: str = "",
        project_description: str = "",
        project_root: str = "",
        project_workspace_name: str = "",
        global_hyfi_root: str = "",
        global_workspace_name: str = "",
        num_workers: int = -1,
        log_level: str = "",
        autotime: bool = True,
        retina: bool = True,
        verbose: Union[bool, int] = False,
        **kwargs,
    ):
        """
        Initialize and start hyfi.

        Args:
                project_name: Name of the project to use.
                task_name: Name of the task to use.
                project_description: Description of the project that will be used.
                project_root: Root directory of the project.
                project_workspace_name: Name of the project's workspace directory.
                global_hyfi_root: Root directory of the global hyfi.
                global_workspace_name: Name of the global hierachical workspace directory.
                num_workers: Number of workers to run.
                log_level: Log level for the log.
                autotime: Whether to automatically set time and / or keep track of run times.
                retina: Whether to use retina or not.
                verbose: Enables or disables logging
        """
        envs = DotEnvConfig(HYFI_VERBOSE=verbose)
        # Set the project name environment variable HYFI_PROJECT_NAME environment variable if project_name is not set.
        if project_name:
            envs.HYFI_PROJECT_NAME = expand_posix_vars(project_name)
        # Set the task name environment variable HYFI_TASK_NAME to the task name.
        if task_name:
            envs.HYFI_TASK_NAME = expand_posix_vars(task_name)
        # Set the project description environment variable HYFI_PROJECT_DESC environment variable.
        if project_description:
            envs.HYFI_PROJECT_DESC = expand_posix_vars(project_description)
        # Set environment variables HYFI_PROJECT_ROOT to the project root if project_root is set to true.
        if project_root:
            envs.HYFI_PROJECT_ROOT = expand_posix_vars(project_root)
        # Set the project workspace name environment variable HYFI_PROJECT_WORKSPACE_NAME environment variable if project_workspace_name is set to the project workspace name.
        if project_workspace_name:
            envs.HYFI_PROJECT_WORKSPACE_NAME = expand_posix_vars(project_workspace_name)
        # Expand the hyfi_root environment variable.
        if global_hyfi_root:
            envs.HYFI_GLOBAL_ROOT = expand_posix_vars(global_hyfi_root)
        # Set the global workspace name environment variable HYFI_GLOBAL_WORKSPACE_NAME environment variable.
        if global_workspace_name:
            envs.HYFI_GLOBAL_WORKSPACE_NAME = expand_posix_vars(global_workspace_name)
        # Set the number of workers to use.
        if num_workers:
            envs.HYFI_NUM_WORKERS = num_workers
        # Set the log level to the given log level.
        if log_level:
            envs.HYFI_LOG_LEVEL = log_level
            setLogger(log_level)
            logger.setLevel(log_level)
        # Load the extentions for the autotime extension.
        if autotime:
            load_extentions(exts=["autotime"])
        # Set the retina matplotlib formats.
        if retina:
            set_matplotlib_formats("retina")
        self.initialize()

    def initialize(self, config: Union[DictConfig, Dict, None] = None):
        """
        Initialize hyfi. This is called by : meth : ` __init__ ` to perform initialisation.

        Args:
                config: Configuration dictionary or None.

        Returns:
                A boolean indicating whether initialization was successful
        """
        """Initialize hyfi config"""
        # Returns the current value of the __initilized__ attribute.
        if self.__initilized__:
            return
        __hydra_config__.hyfi_config_module = self.hyfi_config_module
        __hydra_config__.hyfi_config_path = self.hyfi_config_path
        __hydra_config__.hyfi_user_config_path = self.hyfi_user_config_path

        # If config is not set the default config is used.
        if config is None:
            config = _compose(
                overrides=["+project=__init__"], config_module=__about__.config_module
            )
            logger.debug("Using default config.")

        # Skip project config initialization.
        if "project" not in config:
            logger.info(
                "No project config found, skip project config initialization."
            )
            return
        self.project = ProjectConfig(**config["project"])
        self.project.init_project()
        # Initialize joblib backend if joblib is not set.
        if self.project.joblib:
            self.project.joblib.init_backend()

        self.__initilized__ = True

    def terminate(self):
        """
        Terminate hyfi config by stopping joblib


        Returns:
                True if successful False
        """
        """Terminate hyfi config"""
        # If the module is not initialized yet.
        if not self.__initilized__:
            return
        # Stop the backend if the joblib is running.
        if self.project and self.project.joblib:
            self.project.joblib.stop_backend()
        self.__initilized__ = False

    def __repr__(self):
        """
        Returns a string representation of HyFIConfig.


        Returns:
                The string representation of HyFI
        """
        return f"HyFIConfig(project={self.project})"

    def __str__(self):
        """
        Returns a string representation of the object.


        Returns:
                The string representation of the
        """
        return self.__repr__()

    @property
    def app_version(self):
        """
        Get the version of the application.


        Returns:
                The version of the application
        """
        return self.about.version


__global_config__ = HyfiConfig()



"""Command line interface for HyFI"""

import click

from hyfi.__global__ import __hyfi_path__
from hyfi.__global__.config import HyfiConfig
from hyfi._version import __version__
from hyfi.copier import Copier
from hyfi.main import _about
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.option(
    "--src_path",
    show_default=True,
    default=f"{__hyfi_path__()}/conf",
    help="Source path to copy from",
)
@click.option(
    "--dst_path",
    show_default=True,
    default="./tmp/conf",
    help="Destination path to copy to",
)
@click.option(
    "--exclude",
    show_default=True,
    default=None,
    help="Exclude files matching this pattern",
)
@click.option(
    "--skip_if_exists",
    is_flag=True,
    show_default=True,
    default=False,
    help="Skip if destination exists",
)
@click.option(
    "--overwrite",
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite destination",
)
@click.option(
    "--dry_run", is_flag=True, show_default=True, default=False, help="Dry run"
)
@click.option(
    "--verbose", is_flag=True, show_default=True, default=False, help="Verbose output"
)
def cc(**args):
    """
    Copy all config files to the destination directory.
    """
    click.echo("Copying configuration files")
    # click.echo(f"args : {args}")
    with Copier(**args) as worker:
        worker.run_copy()


@cli.command()
def about():
    """
    Print the about information for Hyfi.
    """
    cfg = HyfiConfig()
    cfg.about.version = __version__
    _about(cfg)


@cli.command()
@click.option(
    "--uninstall",
    "-u",
    is_flag=True,
    show_default=True,
    default=False,
    help="Uninstall shell completion",
)
@click.option(
    "--shell",
    "-s",
    show_default=True,
    default="zsh",
    help="Shell to install completion for",
)
def sc(uninstall, shell):
    """
    Install or Uninstall shell completion for Hyfi.
    """
    if uninstall:
        click.echo(f"Uninstall shell completion for {shell}:")
        click.echo(
            "run the following command to uninstall shell completion in your current shell\n"
        )
        click.echo(f'eval "$(hyfi -sc uninstall={shell})"')
    else:
        click.echo(f"Install shell completion for {shell}:")
        click.echo(
            "run the following command to install shell completion in your current shell\n"
        )
        click.echo(f'eval "$(hyfi -sc install={shell})"')



import random
from pathlib import Path
from typing import Any, Optional, Union

from pydantic import BaseModel, validator

from hyfi.hydra import _compose
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class BatchConfig(BaseModel):
    config_name: str = "__init__"
    batch_name: str
    batch_num: Optional[int] = None
    batch_root: str = "outputs"
    output_suffix: str = ""
    output_extention: Optional[str] = ""
    random_seed: bool = True
    seed: int = -1
    resume_run: bool = False
    resume_latest: bool = False
    device: str = "cpu"
    num_devices: int = 1
    num_workers: int = 1
    config_yaml = "config.yaml"
    config_json = "config.json"
    config_dirname = "configs"
    verbose: Union[bool, int] = False

    def __init__(
        self,
        config_name: str = "__init__",
        **data: Any,
    ):
        data = _compose(
            f"batch={config_name}",
            config_data=data,
        )  # type: ignore
        super().__init__(**data)
        self.init_batch_num()

    def init_batch_num(self):
        if self.batch_num is None:
            num_files = len(list(self.config_dir.glob(self.config_filepattern)))
            self.batch_num = num_files - 1 if self.resume_latest else num_files
        if self.verbose:
            logger.info(
                "Init batch - Batch name: %s, Batch num: %s",
                self.batch_name,
                self.batch_num,
            )

    @validator("seed")
    def _validate_seed(cls, v, values):
        if values["random_seed"] or v is None or v < 0:
            random.seed()
            seed = random.randint(0, 2**32 - 1)
            if values.get("verbose"):
                logger.info(f"Setting seed to {seed}")
            return seed
        return v

    @validator("output_extention")
    def _validate_output_extention(cls, v):
        return v.strip(".") if v else ""

    @property
    def root_dir(self):
        return Path(self.batch_root)

    @property
    def batch_dir(self):
        batch_dir = self.root_dir / self.batch_name
        batch_dir.mkdir(parents=True, exist_ok=True)
        return batch_dir

    @property
    def config_dir(self):
        config_dir = self.batch_dir / self.config_dirname
        config_dir.mkdir(parents=True, exist_ok=True)
        return config_dir

    @property
    def file_prefix(self):
        return f"{self.batch_name}({self.batch_num})"

    @property
    def output_file(self):
        if self.output_suffix:
            return f"{self.file_prefix}_{self.output_suffix}.{self.output_extention}"
        else:
            return f"{self.file_prefix}.{self.output_extention}"

    @property
    def config_filename(self):
        return f"{self.file_prefix}_{self.config_yaml}"

    @property
    def config_jsonfile(self):
        return f"{self.file_prefix}_{self.config_json}"

    @property
    def config_filepattern(self):
        return f"{self.batch_name}(*)_{self.config_yaml}"

    @property
    def config_filepath(self):
        return self.config_dir / self.config_filename

    @property
    def config_jsonpath(self):
        return self.config_dir / self.config_jsonfile



from typing import Any, Dict, List, Union

import hydra
from omegaconf import DictConfig, ListConfig, OmegaConf, SCMode

from hyfi.__global__ import __about__, __hydra_config__, __hydra_version_base__
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def _select(
    cfg: Any,
    key: str,
    *,
    default: Any = None,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
):
    key = key.replace("/", ".")
    return OmegaConf.select(
        cfg,
        key=key,
        default=default,
        throw_on_resolution_failure=throw_on_resolution_failure,
        throw_on_missing=throw_on_missing,
    )


def _to_dict(
    cfg: Any,
) -> Any:
    if isinstance(cfg, dict):
        cfg = _to_config(cfg)
    if isinstance(cfg, (DictConfig, ListConfig)):
        return OmegaConf.to_container(
            cfg,
            resolve=True,
            throw_on_missing=False,
            structured_config_mode=SCMode.DICT,
        )
    return cfg


def _to_config(
    cfg: Any,
) -> Union[DictConfig, ListConfig]:
    return OmegaConf.create(cfg)


def _compose(
    config_group: Union[str, None] = None,
    overrides: Union[List[str], None] = None,
    config_data: Union[Dict[str, Any], DictConfig, None] = None,
    *,
    return_as_dict: bool = True,
    throw_on_resolution_failure: bool = True,
    throw_on_missing: bool = False,
    config_name: Union[str, None] = None,
    config_module: Union[str, None] = None,
    global_package: bool = False,
    verbose: bool = False,
) -> Union[DictConfig, Dict]:  # sourcery skip: low-code-quality
    """
    Compose a configuration by applying overrides

    Args:
        config_group: Name of the config group to compose (`config_group=name`)
        overrides: List of config groups to apply overrides to (`overrides=["override_name"]`)
        config_data: Keyword arguments to override config group values (will be converted to overrides of the form `config_group.key=value`)
        return_as_dict: Return the result as a dict
        throw_on_resolution_failure: If True throw an exception if resolution fails
        throw_on_missing: If True throw an exception if config_group doesn't exist
        config_name: Name of the root config to be used (e.g. `hconf`)
        config_module: Module of the config to be used (e.g. `hyfi.conf`)
        global_package: If True, the config assumed to be a global package
        verbose: If True print configuration to stdout

    Returns:
        A config object or a dictionary with the composed config
    """
    if isinstance(config_data, DictConfig):
        logger.debug("returning config_group_kwargs without composing")
        return (
            _to_dict(config_data)
            if return_as_dict and isinstance(config_data, DictConfig)
            else config_data
        )
    # Set overrides to the empty list if None
    if overrides is None:
        overrides = []
    config_module = config_module or __hydra_config__.hyfi_config_module
    # if verbose:
    logger.debug("config_module: %s", config_module)
    is_initialized = hydra.core.global_hydra.GlobalHydra.instance().is_initialized()  # type: ignore
    # Set the group key and value of the config group.
    if config_group:
        group_ = config_group.split("=")
        # group_key group_value group_key group_value group_key group_value default
        if len(group_) == 2:
            group_key, group_value = group_
        else:
            group_key = group_[0]
            group_value = "default"
        config_group = f"{group_key}={group_value}"
    else:
        group_key = None
        group_value = None
    # If group_key and group_value are specified in the configuration file.
    if group_key and group_value:
        # Initialize hydra configuration module.
        if is_initialized:
            cfg = hydra.compose(config_name=config_name, overrides=overrides)
        else:
            with hydra.initialize_config_module(
                config_module=config_module, version_base=__hydra_version_base__
            ):
                cfg = hydra.compose(config_name=config_name, overrides=overrides)
        cfg = _select(
            cfg,
            key=group_key,
            default=None,
            throw_on_missing=False,
            throw_on_resolution_failure=False,
        )
        override = config_group if cfg is not None else f"+{config_group}"
        # Add override to overrides list.
        if isinstance(override, str):
            if overrides:
                overrides.append(override)
            else:
                overrides = [override]
    # Add config group overrides to overrides list.
    if config_data:
        for k, v in config_data.items():
            if isinstance(v, (str, int)):
                overrides.append(f"{group_key}.{k}={v}")
    # if verbose:
    logger.debug(f"compose config with overrides: {overrides}")
    # Initialize hydra and return the configuration.
    if is_initialized:
        # Hydra is already initialized.
        if verbose:
            logger.debug("Hydra is already initialized")
        cfg = hydra.compose(config_name=config_name, overrides=overrides)
    else:
        with hydra.initialize_config_module(
            config_module=config_module, version_base=__hydra_version_base__
        ):
            cfg = hydra.compose(config_name=config_name, overrides=overrides)

    # Select the group_key from the configuration.
    if group_key and not global_package:
        cfg = _select(
            cfg,
            key=group_key,
            default=None,
            throw_on_missing=throw_on_missing,
            throw_on_resolution_failure=throw_on_resolution_failure,
        )
    logger.debug("Composed config: %s", OmegaConf.to_yaml(_to_dict(cfg)))
    return _to_dict(cfg) if return_as_dict and isinstance(cfg, DictConfig) else cfg



from pydantic import BaseModel

from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class AboutConfig(BaseModel):
    """About Configuration"""

    __package_name__: str = "hyfi"
    name: str = "HyFI"
    authors: str = "Young Joon Lee <entelecheia@hotmail.com>"
    description: str = (
        "Hydra Fast Interface (Hydra and Pydantic based interface framework)"
    )
    homepage: str = "https://hyfi.entelecheia.ai"
    license: str = "MIT"
    version: str = "0.0.0"

    class Config:
        extra = "allow"
        underscore_attrs_are_private = False

    @property
    def config_module(self) -> str:
        return f"{self.__package_name__}.conf"

    @property
    def config_path(self) -> str:
        return f"pkg://{self.config_module}"



"""Utilities for loading libraries and dependencies."""
import importlib
import os
import subprocess
import sys
from pathlib import Path

from hyfi.utils.file import is_dir, is_file
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def gitclone(
    url: str,
    targetdir: str = "",
    verbose: bool = False,
) -> None:
    if targetdir:
        res = subprocess.run(
            ["git", "clone", url, targetdir], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
    else:
        res = subprocess.run(
            ["git", "clone", url], stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        logger.info(res)


def pip(
    name: str,
    upgrade: bool = False,
    prelease: bool = False,
    editable: bool = False,
    quiet: bool = True,
    find_links: str = "",
    requirement: bool = False,
    force_reinstall: bool = False,
    verbose: bool = False,
    **kwargs,
) -> None:
    """Install a package using pip."""
    _cmd = ["pip", "install"]
    if upgrade:
        _cmd.append("--upgrade")
    if prelease:
        _cmd.append("--pre")
    if editable:
        _cmd.append("--editable")
    if quiet:
        _cmd.append("--quiet")
    if find_links:
        _cmd += ["--find-links", find_links]
    if requirement:
        _cmd.append("--requirement")
    if force_reinstall:
        _cmd.append("--force-reinstall")
    for k in kwargs:
        k = k.replace("_", "-")
        _cmd.append(f"--{k}")
    _cmd.append(name)
    if verbose:
        logger.info(f"Installing: {' '.join(_cmd)}")
    res = subprocess.run(_cmd, stdout=subprocess.PIPE).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        logger.info(res)


def pipi(name: str, verbose: bool = False) -> None:
    """Install a package using pip."""
    res = subprocess.run(
        ["pip", "install", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        logger.info(res)


def pipie(name: str, verbose: bool = False) -> None:
    """Install a editable package using pip."""
    res = subprocess.run(
        ["git", "install", "-e", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        logger.info(res)


def apti(name: str, verbose: bool = False) -> None:
    """Install a package using apt."""
    res = subprocess.run(
        ["apt", "install", name], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    if verbose:
        print(res)
    else:
        logger.info(res)


def load_module_from_file(name: str, libpath: str, specname: str = "") -> None:
    """Load a module from a file"""
    module_path = os.path.join(libpath, name.replace(".", os.path.sep))
    if is_file(f"{module_path}.py"):
        module_path = f"{module_path}.py"
    elif is_dir(module_path):
        module_path = os.path.join(module_path, "__init__.py")
    else:
        module_path = str(Path(module_path).parent / "__init__.py")

    spec = importlib.util.spec_from_file_location(name, module_path)  # type: ignore
    module = importlib.util.module_from_spec(spec)  # type: ignore
    if not specname:
        specname = spec.name
    sys.modules[specname] = module
    spec.loader.exec_module(module)


def ensure_import_module(
    name: str, libpath: str, liburi: str, specname: str = "", syspath: str = ""
) -> None:
    """Ensure a module is imported, if not, clone it from a git repo and load it"""
    try:
        if specname:
            importlib.import_module(specname)
        else:
            importlib.import_module(name)
        logger.info(f"{name} imported")
    except ImportError:
        if not os.path.exists(libpath):
            logger.info(f"{libpath} not found, cloning from {liburi}")
            gitclone(liburi, libpath)
        if not syspath:
            syspath = libpath
        if syspath not in sys.path:
            sys.path.append(syspath)
        load_module_from_file(name, syspath, specname)
        specname = specname or name
        logger.info(f"{name} not imported, loading from {syspath} as {specname}")
